#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Optimising hyper-parameters for ML models with Optuna
#


import pandas as pd
import numpy as np
import joblib, json, os, gc

import optuna
from optuna.samplers import RandomSampler, TPESampler


from .util import GroupedTimeSeriesSplit, strategy_metrics
from .benchmark import benchmark_pipeline, save_best_model, load_best_model
from .numerai import load_numerai_data, score_numerai


import logging

logger = logging.getLogger("Numerai")


### Create Hyper-parameter space for optuna
### Extract parameter space that needs to be optimised from the config dictionary
def create_optuna_space(config_dictionary, trial):
    space = dict()
    for step in ["feature_eng", "ml_method"]:
        for k, v in config_dictionary[step]["parameters"].items():
            if isinstance(v, list):
                space[k] = getattr(trial, f"suggest_{v[0]}")(name=k, **v[1])
            else:
                space[k] = v
    return space


### Create Parameter Sets from optuna trial instances
def create_parameters_sets(
    args,
    config_dictionary,
    seed=0,
):
    ### Feature Engineering
    feature_eng_parameters = {}
    for k, v in config_dictionary["feature_eng"]["parameters"].items():
        feature_eng_parameters[k] = args.get(k, v)

    ### ML Methods
    tabular_hyper = {
        "seed": seed,
    }

    for k, v in config_dictionary["ml_method"]["parameters"].items():
        tabular_hyper[k] = args.get(k, v)

    ### Additional Hyper-parameters to be passed to training loop, NOT used now
    additional_hyper = dict()

    return feature_eng_parameters, tabular_hyper, additional_hyper


# Create Objective function using optuna for Numerai Classic and Numerai Signals Tournament


def create_optuna_numerai_objective(
    config_dictionary, numerai_files, seed=0, debug=False
):
    def objective(trial):
        with open(numerai_files["feature_metadata"], "r") as f:
            feature_metadata = json.load(f)
        if config_dictionary["model_params"]["feature_sets"] == "v4":
            features_optimizer = feature_metadata["feature_sets"]["fncv3_features"]
            bad_features = [
                "feature_palpebral_univalve_pennoncel",
                "feature_unsustaining_chewier_adnoun",
                "feature_brainish_nonabsorbent_assurance",
                "feature_coastal_edible_whang",
                "feature_disprovable_topmost_burrower",
                "feature_trisomic_hagiographic_fragrance",
                "feature_queenliest_childing_ritual",
                "feature_censorial_leachier_rickshaw",
                "feature_daylong_ecumenic_lucina",
                "feature_steric_coxcombic_relinquishment",
            ]
            features_optimizer = list(set(features_optimizer) - set(bad_features))
        else:
            features_optimizer = list()

        features, targets, groups, weights = load_numerai_data(
            numerai_files["dataset"],
            feature_metadata=numerai_files["feature_metadata"],
            resample=0,
            resample_freq=config_dictionary["model_params"]["train_resample_freq"],
            target_col=config_dictionary["model_params"]["train_targets"],
            data_version=config_dictionary["model_params"]["feature_sets"],
            startera=config_dictionary["model_params"]["train_startera"],
            endera=config_dictionary["model_params"]["train_endera"],
        )

        param = create_optuna_space(config_dictionary, trial)

        logger.info(param)

        (
            feature_eng_parameters,
            tabular_hyper,
            additional_hyper,
        ) = create_parameters_sets(
            param,
            config_dictionary,
            seed=seed,
        )

        model_performance, trained_models, data, parameters = benchmark_pipeline(
            features,
            targets,
            weights,
            groups,
            feature_eng=config_dictionary["feature_eng"]["method"],
            feature_eng_parameters=feature_eng_parameters,
            tabular_model=config_dictionary["ml_method"]["method"],
            tabular_hyper=tabular_hyper,
            model_params=config_dictionary["model_params"]["train"],
            additional_hyper=additional_hyper,
            debug=debug,
        )
        ## Get Predictions for each of the walk forward model
        ## Score on Validation data
        predictions = list()
        for model_name in list(data.keys()):
            predictions.append(data[model_name]["prediction"])
        train_prediction_df = pd.DataFrame(pd.concat(predictions, axis=0).mean(axis=1))
        train_prediction_df.columns = ["prediction"]
        train_prediction_df["target"] = targets.reindex(train_prediction_df.index)
        train_prediction_df["era"] = groups.reindex(train_prediction_df.index)
        train_prediction_df, correlations_by_era = score_numerai(
            train_prediction_df,
            features,
            riskiest_features=features_optimizer,
            proportion=float(
                config_dictionary["model_params"]["selection"]["proportion"]
            ),
            era_col="era",
            target_col_name="target",
        )
        performances = strategy_metrics(correlations_by_era["neutralised_correlation"])
        metric = performances[
            config_dictionary["model_params"]["selection"]["criteria"]
        ]
        logger.info(f"Out of Sample Metric {metric}")
        return metric

    return objective


def optuna_search(
    config_dictionary,
    numerai_files,
    n_trials=10,
    timeout=10000,
    seed=0,
    debug=False,
):
    optuna.logging.set_verbosity(optuna.logging.WARNING)

    numerai_objective = create_optuna_numerai_objective(
        config_dictionary, numerai_files, seed=seed, debug=debug
    )
    study = optuna.create_study(
        direction="maximize",
    )
    study.optimize(
        numerai_objective, n_trials=n_trials, timeout=timeout, gc_after_trial=True
    )

    return study.best_trial.params, study.best_trial.value


def train_best_model_optuna(
    target_col_name,
    end_era,
    best_parameters,
    config_dictionary,
    numerai_files,
    seed=0,
    debug=False,
):
    resample_seed = int(
        seed % config_dictionary["model_params"]["validate_resample_freq"]
    )
    features, targets, groups, weights = load_numerai_data(
        numerai_files["dataset"],
        feature_metadata=numerai_files["feature_metadata"],
        resample=resample_seed,
        resample_freq=config_dictionary["model_params"]["validate_resample_freq"],
        target_col=[target_col_name],
        data_version=config_dictionary["model_params"]["feature_sets"],
        startera=config_dictionary["model_params"]["train_startera"],
        endera=end_era,
    )

    output_folder = config_dictionary["model_params"]["output_folder"]

    if not os.path.exists(f"{output_folder}/"):
        os.mkdir(f"{output_folder}/")

    feature_eng_parameters, tabular_hyper, additional_hyper = create_parameters_sets(
        best_parameters,
        config_dictionary,
        seed=seed,
    )

    model_performance, trained_models, data, parameters = benchmark_pipeline(
        features,
        targets,
        weights,
        groups,
        feature_eng=config_dictionary["feature_eng"]["method"],
        feature_eng_parameters=feature_eng_parameters,
        tabular_model=config_dictionary["ml_method"]["method"],
        tabular_hyper=tabular_hyper,
        model_params=config_dictionary["model_params"]["validate"],
        additional_hyper=additional_hyper,
        debug=debug,
    )

    ## Save each model
    for model_name in list(trained_models.keys()):
        ## Save Parameters and Feature Transformer
        output_parameters_path = f"{output_folder}/{model_name}_{seed}.parameters"
        output_parameters = dict()
        output_parameters["parameters"] = parameters[model_name]
        output_parameters["transformer"] = trained_models[model_name]["transformer"]
        joblib.dump(output_parameters, output_parameters_path)

        ## Save Model
        output_model_path = f"{output_folder}/{model_name}_{seed}.model"
        save_best_model(
            trained_models[model_name]["model"],
            parameters[model_name]["model"]["tabular_model"],
            output_model_path,
        )

    return None


def numerai_optimisation_pipeline_optuna(
    config_dictionary,
    numerai_files,
    run_optimisation=True,
    optimised_parameters_path="numerai_best_parameters.json",
    grid_search_seed=0,
    n_trials=40,
    timeout=2000,
    debug=False,
):
    ## Search for optimal hyper-parameters
    if run_optimisation:
        best_parameters, best_value = optuna_search(
            config_dictionary,
            numerai_files,
            seed=grid_search_seed,
            n_trials=n_trials,
            timeout=timeout,
            debug=debug,
        )
        with open(optimised_parameters_path, "w") as f:
            best_parameters["Optuna_Best_Value"] = best_value
            json.dump(best_parameters, f)
    else:
        with open(optimised_parameters_path, "r") as f:
            best_parameters = json.load(f)
            logger.info(f"Using Best parameters {best_parameters}")

    START_SEED = config_dictionary["model_params"]["model_no_start"]
    NO_MODELS_PER_CONFIG = config_dictionary["model_params"]["no_models_per_config"]

    if config_dictionary["model_params"]["mix_cv"]:
        for target_col_name in config_dictionary["model_params"]["validate_targets"]:
            for end_era in config_dictionary["model_params"]["validate_enderas"]:
                for seed in range(START_SEED, START_SEED + NO_MODELS_PER_CONFIG):
                    ## Check if Model already exists
                    output_folder = config_dictionary["model_params"]["output_folder"]
                    tabular_model = config_dictionary["ml_method"]["method"]
                    feature_eng = config_dictionary["feature_eng"]["method"]
                    model_name = "{}_{}_{}".format(tabular_model, feature_eng, 1)
                    output_model_path = f"{output_folder}/{model_name}_{seed}.model"
                    if not os.path.exists(output_model_path):
                        train_best_model_optuna(
                            target_col_name,
                            end_era,
                            best_parameters,
                            config_dictionary,
                            numerai_files,
                            seed=seed,
                            debug=debug,
                        )

                START_SEED = START_SEED + NO_MODELS_PER_CONFIG
    else:
        for end_era in config_dictionary["model_params"]["validate_enderas"]:
            for target_col_name in config_dictionary["model_params"][
                "validate_targets"
            ]:
                for seed in range(START_SEED, START_SEED + NO_MODELS_PER_CONFIG):
                    ## Check if Model already exists
                    output_folder = config_dictionary["model_params"]["output_folder"]
                    tabular_model = config_dictionary["ml_method"]["method"]
                    feature_eng = config_dictionary["feature_eng"]["method"]
                    model_name = "{}_{}_{}".format(tabular_model, feature_eng, 1)
                    output_model_path = f"{output_folder}/{model_name}_{seed}.model"
                    if not os.path.exists(output_model_path):
                        train_best_model_optuna(
                            target_col_name,
                            end_era,
                            best_parameters,
                            config_dictionary,
                            numerai_files,
                            seed=seed,
                            debug=debug,
                        )

                START_SEED = START_SEED + NO_MODELS_PER_CONFIG



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# A collection of tools for data processing for Numerai and other temporal tabular data competitions
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import joblib, datetime, json, os
import pandas as pd
import numpy as np
import scipy
import torch
import xgboost

from .benchmark import load_best_model
from .util import strategy_metrics
from .feature import NumeraiTransformer

if torch.cuda.is_available():
    import cupy as cp
    import cudf
    from cuml.neighbors import KNeighborsRegressor
else:
    from sklearn.neighbors import KNeighborsRegressor


"""
Helper Functions to convert Numerai Era and Datetime 
"""


## Shifting Numerai Era
def shift_era(era, gap=6):
    new_era_int = int(era) + gap
    new_era = str(new_era_int)
    while len(new_era) < 4:
        new_era = "0" + new_era
    return new_era


## Convert datetime into Numerai eras
def convert_datetime_to_era(sample_date):
    baseline = datetime.datetime(year=2003, month=1, day=3)
    differences = datetime.datetime.strptime(sample_date, "%Y-%m-%d") - baseline
    new_era = str(differences.days // 7 + 1)
    while len(new_era) < 4:
        new_era = "0" + new_era
    return new_era


def convert_era_to_datetime(era):
    baseline = datetime.datetime(year=2003, month=1, day=3)
    new_datetime = baseline + datetime.timedelta(days=7 * (int(era) - 1))
    return new_datetime


### Map columns Numerai Era to datetime
def create_era_index(
    df,
    baseline=datetime.datetime(year=2003, month=1, day=3),
):
    mapped_era = [
        baseline + datetime.timedelta(days=7 * (int(x) - 1)) for x in df.index
    ]
    df.index = mapped_era
    return df


"""
Data Loader for Numerai Data 

"""


def load_numerai_data_era(
    filename,
    feature_metadata="v4_features.json",
    resample=0,
    resample_freq=1,
    target_col=["target"],
    era_col="era",
    data_version="v4",
    startera=None,
    endera=None,
):
    ## Read Train Data
    df_raw = pd.read_parquet(filename)
    ## Select Range
    if startera is not None and endera is not None:
        df_raw = df_raw[(df_raw[era_col] <= endera) & (df_raw[era_col] >= startera)]
    elif endera is not None:
        df_raw = df_raw[(df_raw[era_col] <= endera)]
    ## Downsample Eras
    if resample_freq > 1:
        downsampled_eras = df_raw[era_col].unique()[resample::resample_freq]
        df = df_raw[df_raw[era_col].isin(downsampled_eras)]
    else:
        df = df_raw.copy()

    del df_raw

    ## Features Sets
    feature_col = [col for col in df.columns if col.startswith("feature_")]

    if data_version in [
        "v4",
        "v4-all",
    ]:
        bad_features = [
            "feature_palpebral_univalve_pennoncel",
            "feature_unsustaining_chewier_adnoun",
            "feature_brainish_nonabsorbent_assurance",
            "feature_coastal_edible_whang",
            "feature_disprovable_topmost_burrower",
            "feature_trisomic_hagiographic_fragrance",
            "feature_queenliest_childing_ritual",
            "feature_censorial_leachier_rickshaw",
            "feature_daylong_ecumenic_lucina",
            "feature_steric_coxcombic_relinquishment",
        ]
        feature_col = list(set(feature_col) - set(bad_features))

    ## Features and Targets are DataFrame
    if data_version == "signals":
        features = df[feature_col].fillna(0)
    ## For Numerai Classic Tournament, v4 dataset
    else:
        features = df[feature_col].fillna(2) - 2

    target_median = df[target_col].median()
    targets = df[target_col].fillna(target_median) - target_median
    ## Group column has to be pd.Series for time-series cross validation
    groups = df[era_col]
    ## weights column has to be pd.Series for time-series cross validation
    df["weights"] = 1
    weights = df["weights"]
    return features.astype(np.int8), targets.astype(np.float32), groups, weights


def load_numerai_data(
    data_folder,
    feature_metadata="v4_features.json",
    resample=0,
    resample_freq=1,
    target_col=["target"],
    era_col="era",
    data_version="v4",
    startera=None,
    endera=None,
):
    if data_version in [
        "v4",
        "v4.1",
        "v5",
        "v6",
    ]:
        features_list = list()
        targets_list = list()
        groups_list = list()
        weights_list = list()

        if startera is None:
            startera = "0001"
        if endera is None:
            endera = "0001"

        for i in range(int(startera) + resample, int(endera) + 1, resample_freq):
            if i <= 9:
                test_start_str = "000" + str(i)
            elif i <= 99:
                test_start_str = "00" + str(i)
            elif i <= 999:
                test_start_str = "0" + str(i)
            else:
                test_start_str = str(i)

            data_file = f"{data_folder}/{data_version}_{test_start_str}_int8.parquet"

            features, targets, groups, weights = load_numerai_data_era(
                data_file,
                feature_metadata=feature_metadata,
                resample=0,
                resample_freq=1,
                target_col=target_col,
                era_col=era_col,
                data_version=data_version,
                startera=test_start_str,
                endera=test_start_str,
            )

            features_list.append(features)
            targets_list.append(targets)
            groups_list.append(groups)
            weights_list.append(weights)

        return (
            pd.concat(features_list),
            pd.concat(targets_list),
            pd.concat(groups_list),
            pd.concat(weights_list),
        )
    else:
        features, targets, groups, weights = load_numerai_data_era(
            data_folder,
            feature_metadata=feature_metadata,
            resample=resample,
            resample_freq=resample_freq,
            target_col=target_col,
            era_col=era_col,
            data_version=data_version,
            startera=startera,
            endera=endera,
        )
        return features, targets, groups, weights


"""
Generate Predictions for Numerai 

trained_model: model object which has method .predict to generate predictions
parameters: dictionary which contains parameters of the trained_model
modelname: str Name of Model
start_iteration: for tree-based methods, skip the first N trees in model when generating predictions 
startera: first era to get predictions 
endera: last era to get predictions 


Output: prediction_df: pd.DataFrame with columns era, prediction, model_name, target_col 

"""


class FeatureMomentumModel:
    def __init__(
        self,
        lookback=52,
        shift=6,
        correlation_file_path=None,
        portfolio_file_path=None,
        target_col=None,
        seed=0,
    ):
        self.seed = seed
        self.lookback = lookback
        self.shift = shift
        self.correlation_file_path = correlation_file_path
        self.portfolio_file_path = portfolio_file_path
        self.target_col = target_col

    def predict(self, features):
        correlation_matrix = pd.read_parquet(self.correlation_file_path)
        factor_momentum = (
            correlation_matrix.shift(self.shift)
            .fillna(0)
            .rolling(self.lookback)
            .mean()
            .dropna()
        )
        last_momentum = factor_momentum.tail(1).transpose().squeeze()[features.columns]
        preds = features * np.sign(last_momentum)
        return preds.mean(axis=1)

    def copy_performance(self, outputfolder):
        portfolio = pd.read_csv(self.portfolio_file_path, index_col=0)
        portfolio.columns = [f"feature-momentum_None_1_{self.seed}_1-baseline-0"]
        portfolio.to_csv(f"{outputfolder}/feature-momentum_None_1_{self.seed}_1.csv")


def predict_numerai(
    features_raw,
    targets,
    groups,
    trained_model,
    parameters,
    modelname="sample",
    gbm_start_iteration=0,  ## Backward Comptability
    era_col="era",
    debug=False,
):
    ## Score on Dataset

    selected_cols = parameters["parameters"]["model"]["feature_columns"]
    target_col = parameters["parameters"]["model"]["target_columns"]

    ## Transform Features
    if parameters["parameters"]["model"]["feature_engineering"] is not None:
        if parameters["parameters"]["model"]["feature_engineering"] in [
            "numerai",
        ]:
            feature_eng_parameters = parameters["parameters"]["feature_eng"]
            transformer = NumeraiTransformer(**feature_eng_parameters)
            transformer.data = parameters["transformer"]
            features = transformer.transform(
                features_raw[selected_cols], is_train=False
            )
        if parameters["parameters"]["model"]["feature_engineering"] in [
            "numeraiv4",
            "numeraiv4.1",
        ]:
            feature_eng_parameters = parameters["parameters"]["feature_eng"]
            transformer = NumeraiTransformerV4(**feature_eng_parameters)
            transformer.data = parameters["transformer"]
            features = transformer.transform(
                features_raw[selected_cols], is_train=False
            )
    else:
        features = features_raw[selected_cols]

    ## Run Predictions
    ## For tree-based models can run some of the trees only
    if parameters["parameters"]["model"]["tabular_model"] in [
        "lightgbm",
        "lightgbm-gbdt",
        "lightgbm-dart",
        "lightgbm-goss",
    ]:
        ## Backward Compatability
        if "additional" in parameters["parameters"] and gbm_start_iteration is None:
            gbm_start_iteration = parameters["parameters"]["additional"].get(
                "gbm_start_iteration", 0
            )
        start_iteration = min(
            gbm_start_iteration, int(trained_model.num_trees() * 0.75)
        )
        predictions_raw = trained_model.predict(
            features, start_iteration=start_iteration
        )
    elif parameters["parameters"]["model"]["tabular_model"] in [
        "xgboost",
    ]:
        if hasattr(trained_model, "best_iteration"):
            end_iteration = trained_model.best_iteration
        else:
            end_iteration = trained_model.num_boosted_rounds()
        start_iteration = min(gbm_start_iteration, int(end_iteration * 0.75))
        xgboost_features = xgboost.DMatrix(features)
        predictions_raw = trained_model.predict(
            xgboost_features,
            iteration_range=(start_iteration, end_iteration),
        )
    elif parameters["parameters"]["model"]["tabular_model"] in [
        "Numerai-MLP",
        "Numerai-LSTM",
    ]:
        predictions_raw = trained_model.predict(features.values)
    elif parameters["parameters"]["model"]["tabular_model"] in [
        "tabnet",
    ]:
        predictions_raw = trained_model.predict(features.values)
    elif parameters["parameters"]["model"]["tabular_model"] in [
        "feature-momentum",
    ]:
        trained_model = FeatureMomentumModel(**parameters["parameters"]["tabular"])
        predictions_raw = trained_model.predict(features)
    else:
        ## General Model which implements a predict method
        predictions_raw = trained_model.predict(features)

    ## Process Predictions into DataFrame
    predictions = pd.DataFrame(
        predictions_raw,
        columns=target_col,
        index=targets.index,
    )
    predictions[era_col] = groups
    ## Rank Predictions within each era
    normalised_predictions = list()
    for i, df in predictions.groupby(era_col):
        per_era = df[target_col].rank(pct=True, axis=0)
        normalised_predictions.append(per_era)
    processed_predictions = pd.concat(normalised_predictions, axis=0)
    predictions["prediction"] = processed_predictions[target_col].mean(axis=1)
    prediction_df = pd.concat([predictions[[era_col, "prediction"]], targets], axis=1)
    prediction_df["model_name"] = modelname
    return prediction_df


def predict_numerai_multiple(
    Numerai_Model_Names,
    correlation_matrix=None,
    filename="data/v4_all_int8.parquet",
    data_version="v4",
    startera=None,
    endera=None,
    debug=False,
    era_col="era",
    target_col=["target"],
    embargo=26,
    gbm_start_iteration=0,
):
    features, targets, groups, weights = load_numerai_data(
        filename,
        target_col=target_col,
        era_col=era_col,
        data_version=data_version,
        startera=startera,
        endera=endera,
    )

    INDEX_COL_NAMES = features.index.names

    prediction_df_list = list()
    score_df_list = list()

    for Numerai_Model_Name in Numerai_Model_Names:
        modelname = Numerai_Model_Name.replace(".parameters", ".model")
        parameters = joblib.load(Numerai_Model_Name)
        most_recent_model = load_best_model(
            parameters["parameters"]["model"]["tabular_model"], modelname
        )

        ## Check Embargo Period for Numerai Classic Models
        if data_version in [
            "v4",
            "v4.1",
            "v5",
            "v6",
        ]:
            test_start = shift_era(
                parameters["parameters"]["model"]["validation_end"], embargo
            )
            required_index = groups[groups >= test_start].index
        else:
            required_index = groups.index

        if debug:
            print(modelname, test_start)

        if required_index.shape[0] > 0:
            prediction_df = predict_numerai(
                features.loc[required_index],
                targets.loc[required_index],
                groups.loc[required_index],
                most_recent_model,
                parameters,
                modelname=modelname,
                gbm_start_iteration=gbm_start_iteration,
                era_col=era_col,
                debug=debug,
            )
            prediction_df_list.append(prediction_df)

            if debug:
                print(prediction_df.columns, prediction_df.shape)

    if len(prediction_df_list) > 0:
        output_cols = [era_col, "prediction"] + target_col
        average_prediction_df = (
            pd.concat(prediction_df_list, axis=0)
            .groupby(INDEX_COL_NAMES)[output_cols]
            .mean()
        )
        average_prediction_df[era_col] = groups
        if debug:
            print(average_prediction_df.columns, average_prediction_df.shape)

        return average_prediction_df.sort_values(era_col), prediction_df_list
    else:
        return pd.DataFrame(), pd.DataFrame()


"""
Score Numerai Models with FN using CUDA 

prediction_df: pd.DataFrame with columns era, prediction, model_name, target_col  and index id 
features: pd.DataFrame with columns feature_xxx and index id 
riskiest_fatures: list of str 


Output 
prediction_df: pd.DataFrame with columns era, model_name, prediction, neutralised_prediction, target_col, index id 
correlations_by_era: pd.DataFrame with columns correlation, normalised_correlation, neutralised_correlation, index era 

"""


def score_numerai(
    prediction_df,
    features,
    riskiest_features,
    proportion=0,
    modelname="sample",
    target_col_name="target",
    prediction_col="prediction",
    era_col="era",
    debug=False,
):
    ## Find Correlation by era
    correlations_by_era = list()
    for i, df in prediction_df.groupby(era_col):
        output = dict()
        output[era_col] = i
        ## Computation on CUDA
        if torch.cuda.is_available():
            temp = (
                scipy.stats.rankdata(df[prediction_col], method="ordinal") - 0.5
            ) / len(df[prediction_col])
            df["normalised_prediction"] = scipy.stats.norm.ppf(temp)
            ## Neutralised targets (FNC)
            if proportion > 0 and len(riskiest_features) > 0:
                exposures = cp.asarray(features.loc[df.index, riskiest_features])
                normalised_prediction = cp.asarray(df["normalised_prediction"])
                gram_mtx = cp.dot(cp.linalg.pinv(exposures), normalised_prediction)
                projected_values = normalised_prediction - cp.asarray(
                    proportion
                ) * cp.dot(exposures, gram_mtx)
                df["neutralised_prediction"] = projected_values.get()
                df["neutralised_prediction"] = (
                    df["neutralised_prediction"] / df["neutralised_prediction"].std()
                )
                output["neutralised_correlation"] = cp.corrcoef(
                    cp.asarray(df[target_col_name]),
                    cp.asarray(df["neutralised_prediction"].rank(pct=True)),
                )[0, 1].get()
                prediction_df.loc[df.index, "neutralised_prediction"] = df[
                    "neutralised_prediction"
                ].rank(pct=True)
            else:
                output["neutralised_correlation"] = cp.corrcoef(
                    cp.asarray(df[target_col_name]),
                    cp.asarray(df[prediction_col].rank(pct=True)),
                )[0, 1].get()
                prediction_df.loc[df.index, "neutralised_prediction"] = df[
                    prediction_col
                ].rank(pct=True)
        ### Computation on CPU
        else:
            ## Normalise prediction
            temp = (
                scipy.stats.rankdata(df[prediction_col], method="ordinal") - 0.5
            ) / len(df[prediction_col])
            df["normalised_prediction"] = scipy.stats.norm.ppf(temp)
            ## Neutralised targets (FNC)
            if proportion > 0 and len(riskiest_features) > 0:
                exposures = features.loc[df.index, riskiest_features]
                df["neutralised_prediction"] = df[
                    "normalised_prediction"
                ] - proportion * exposures.dot(
                    np.linalg.pinv(exposures).dot(df["normalised_prediction"])
                )
                df["neutralised_prediction"] = (
                    df["neutralised_prediction"] / df["neutralised_prediction"].std()
                )
                output["neutralised_correlation"] = np.corrcoef(
                    df[target_col_name], df["neutralised_prediction"].rank(pct=True)
                )[0, 1]
                prediction_df.loc[df.index, "neutralised_prediction"] = df[
                    "neutralised_prediction"
                ].rank(pct=True)
            else:
                output["neutralised_correlation"] = np.corrcoef(
                    df[target_col_name], df[prediction_col].rank(pct=True)
                )[0, 1]
                prediction_df.loc[df.index, "neutralised_prediction"] = df[
                    prediction_col
                ].rank(pct=True)
        correlations_by_era.append(output)
    ## Generate Overall files
    correlations_by_era_all = pd.DataFrame.from_records(correlations_by_era)
    prediction_df["model_name"] = modelname
    correlations_by_era_all["model_name"] = modelname
    return prediction_df, correlations_by_era_all


"""
Linear Factor Model
Factor Timing
rawdata: pd.DataFrame: Numerai dataset with columns containing the 1149 features and 20 targets, index id 
"""


def numerai_feature_correlation_matrix(
    rawdata, feature_col=None, target_col_name=None, era_col="era"
):
    output = dict()
    for i, df in rawdata.groupby(era_col):
        corr_dict = dict()
        for feature in feature_col:
            corr_dict[feature] = np.corrcoef(
                df[feature].fillna(2).astype(float), df[target_col_name]
            )[0, 1]
        output[i] = corr_dict

    return pd.DataFrame.from_records(output).transpose()[feature_col]


def numerai_feature_momentum(
    data_folder="../data/era",
    output_folder="../data/feature_momentum",
    data_version="v4",
    startera="0001",
    endera="1037",
    era_col="era",
    lookback=52,
    update_correlation_mtx=True,
    feature_col=None,
):
    if update_correlation_mtx:
        ## Calculate Correlation Matrix
        for i in range(int(startera), int(endera) + 1):
            if i <= 9:
                test_start_str = "000" + str(i)
            elif i <= 99:
                test_start_str = "00" + str(i)
            elif i <= 999:
                test_start_str = "0" + str(i)
            else:
                test_start_str = str(i)
            data_file = f"{data_folder}/{data_version}_{test_start_str}_int8.parquet"
            rawdata = pd.read_parquet(data_file)
            if feature_col is None:
                feature_col = [x for x in rawdata.columns if x.startswith("feature_")]
            target_cols = [x for x in rawdata.columns if x.startswith("target_")]
            for target_col in target_cols:
                correlation_file = f"{output_folder}/{target_col}_corr.parquet"
                if os.path.exists(correlation_file):
                    correlation_matrix_old = pd.read_parquet(correlation_file)
                    if test_start_str > correlation_matrix_old.index[-1]:
                        rawdata_copy = rawdata.dropna(subset=[target_col]).copy()
                        feature_col = correlation_matrix_old.columns
                        if rawdata_copy.shape[0] > 0:
                            correlation_matrix = numerai_feature_correlation_matrix(
                                rawdata_copy, feature_col, target_col
                            )
                            pd.concat(
                                [correlation_matrix_old, correlation_matrix]
                            ).to_parquet(correlation_file)
                else:
                    rawdata_copy = rawdata.dropna(subset=[target_col]).copy()
                    if rawdata_copy.shape[0] > 0:
                        correlation_matrix = numerai_feature_correlation_matrix(
                            rawdata_copy, feature_col, target_col
                        )
                        correlation_matrix.to_parquet(correlation_file)

    data_file = f"{data_folder}/{data_version}_{endera}_int8.parquet"
    rawdata = pd.read_parquet(data_file)
    # feature_col = [x for x in rawdata.columns if x.startswith("feature_")]
    target_cols = [x for x in rawdata.columns if x.startswith("target_")]

    ## Factor Momentum Portfolio
    for target_col in target_cols:
        correlation_file = f"{output_folder}/{target_col}_corr.parquet"
        correlation_matrix = pd.read_parquet(correlation_file)
        feature_col = correlation_matrix.columns

        if "60" in target_col:
            gap = 14
        else:
            gap = 6

        factor_momentum = (
            correlation_matrix.shift(gap).fillna(0).rolling(lookback).mean().dropna()
        )
        factor_volatility = (
            correlation_matrix.shift(gap).fillna(0).rolling(lookback).std().dropna()
        )
        fm_max_index = factor_momentum.index.max()
        fm_min_index = factor_momentum.index.min()

        factor_momentum_eras = factor_momentum.unstack(level=0).reset_index()
        factor_momentum_eras.columns = ["feature_name", "era", "momentum"]
        factor_volatility_eras = factor_volatility.unstack(level=0).reset_index()
        factor_volatility_eras.columns = ["feature_name", "era", "volatility"]

        for i in range(int(startera), int(endera) + 1):
            if i <= 9:
                test_start_str = "000" + str(i)
            elif i <= 99:
                test_start_str = "00" + str(i)
            elif i <= 999:
                test_start_str = "0" + str(i)
            else:
                test_start_str = str(i)

            if (test_start_str <= fm_max_index) & (test_start_str >= fm_min_index):
                factor_file = f"{output_folder}/{target_col}_feature_momentum.csv"
                if os.path.exists(factor_file):
                    factor_portfolio_old = pd.read_csv(factor_file, index_col=0)
                    factor_portfolio_old.index = pd.to_datetime(
                        factor_portfolio_old.index
                    )
                    if (
                        convert_era_to_datetime(test_start_str)
                        > factor_portfolio_old.index[-1]
                    ):
                        update = True
                    else:
                        update = False
                else:
                    update = True

                if update:
                    ## Read Data
                    data_file = (
                        f"{data_folder}/{data_version}_{test_start_str}_int8.parquet"
                    )
                    df = pd.read_parquet(data_file)
                    feature_col = [x for x in df.columns if x.startswith("feature_")]
                    df[feature_col] = df[feature_col].fillna(2) - 2

                    ## Factor Momentum
                    portfolio_predictions = df[[era_col, "target"]]
                    per_era = df[feature_col] * np.sign(
                        factor_momentum.loc[test_start_str, feature_col]
                    )
                    portfolio_predictions["prediction"] = per_era.mean(axis=1)
                    prediction_era, correlations_era = score_numerai(
                        portfolio_predictions,
                        df[feature_col],
                        None,
                        proportion=0,
                        modelname=f"{target_col}_feature_momentum-baseline-0",
                        target_col_name="target",
                    )
                    factor_porfolio = create_era_index(
                        correlations_era.pivot(
                            index="era",
                            columns=["model_name"],
                            values=["neutralised_correlation"],
                        )
                    )
                    if os.path.exists(factor_file):
                        factor_portfolio_old = pd.read_csv(factor_file, index_col=0)
                        factor_portfolio_old.index = pd.to_datetime(
                            factor_portfolio_old.index
                        )
                        pd.concat(
                            [
                                factor_portfolio_old,
                                factor_porfolio["neutralised_correlation"],
                            ]
                        ).to_csv(factor_file)
                    else:
                        factor_porfolio["neutralised_correlation"].to_csv(factor_file)


"""

Benchmark Performances of Numerai Models 

Run Model Performances for models trained with a single ML model 

"""


def dynamic_feature_neutralisation(
    prediction_df,
    features_raw,
    feature_corr=None,
    features_optimizer=None,
    modelname="sample",
    era_col="era",
    target_col=["target"],
    cutoff=420,
    gap=6,
    lookback=52,
    proportion=1,
    debug=False,
):
    if features_optimizer is None:
        features_optimizer = features_raw.columns[:cutoff]

    if feature_corr is None:
        ## Get index by era
        prediction_dynamic = list()
        correlation_dynamic = list()
        for i, df in prediction_df.groupby(era_col):
            if debug:
                print(modelname, i, df.shape)
            prediction_df_era = prediction_df.loc[df.index]
            features_raw_era = features_raw.loc[df.index]
            ## Baseline
            prediction_df_era_new, correlations_by_era = score_numerai(
                prediction_df_era,
                features_raw_era,
                list(),
                proportion=0,
                modelname=f"{modelname}-baseline",
                target_col_name=target_col[0],
                era_col=era_col,
                debug=debug,
            )
            prediction_dynamic.append(prediction_df_era_new.copy())
            correlation_dynamic.append(correlations_by_era)
        return pd.concat(prediction_dynamic, axis=0), pd.concat(
            correlation_dynamic, axis=0
        )

    else:
        ## Generate Feature Momentum Leaderboard
        factor_mean = (
            feature_corr.shift(gap).fillna(0).rolling(lookback).mean().dropna()
        )
        factor_volatility = (
            feature_corr.shift(gap).fillna(0).rolling(lookback).std().dropna()
        )
        factor_skew = (
            feature_corr.shift(gap).fillna(0).rolling(lookback).skew().dropna()
        )
        factor_kurt = (
            feature_corr.shift(gap).fillna(0).rolling(lookback).kurt().dropna()
        )
        factor_drawdown = (
            (-1 * (feature_corr.cumsum() - feature_corr.cumsum().cummax()).cummin())
            .shift(gap)
            .fillna(0)
        )
        factor_sharpe = factor_mean / factor_volatility
        factor_calmar = factor_mean / factor_drawdown
        factor_autocorrelation = (
            feature_corr.rolling(lookback)
            .corr(feature_corr.shift(4))
            .shift(gap)
            .fillna(0)
        )

        fm_max_index = factor_mean.index.max()
        fm_min_index = factor_mean.index.min()

        ##
        factor_flavour_eras = dict()
        for flavour in [
            "mean",
            "volatility",
        ]:
            factor_flavour_eras[flavour] = (
                locals()[f"factor_{flavour}"].unstack(level=0).reset_index()
            )
            factor_flavour_eras[flavour].columns = ["feature_name", "era", flavour]

        ## Get index by era
        prediction_dynamic = list()
        correlation_dynamic = list()
        for i, df in prediction_df.groupby(era_col):
            if debug:
                print(modelname, i, df.shape)
            if (i <= fm_max_index) & (i >= fm_min_index):
                prediction_df_era = prediction_df.loc[df.index]
                features_raw_era = features_raw.loc[df.index]

                ## Baseline
                prediction_df_era_new, correlations_by_era = score_numerai(
                    prediction_df_era,
                    features_raw_era,
                    list(),
                    proportion=0,
                    modelname=f"{modelname}-baseline",
                    target_col_name=target_col[0],
                    era_col=era_col,
                    debug=debug,
                )
                prediction_dynamic.append(prediction_df_era_new.copy())
                correlation_dynamic.append(correlations_by_era)

                ## For v4-data only
                bad_features = [
                    "feature_palpebral_univalve_pennoncel",
                    "feature_unsustaining_chewier_adnoun",
                    "feature_brainish_nonabsorbent_assurance",
                    "feature_coastal_edible_whang",
                    "feature_disprovable_topmost_burrower",
                    "feature_trisomic_hagiographic_fragrance",
                    "feature_queenliest_childing_ritual",
                    "feature_censorial_leachier_rickshaw",
                    "feature_daylong_ecumenic_lucina",
                    "feature_steric_coxcombic_relinquishment",
                ]

                features_optimizer = list(set(features_optimizer) - set(bad_features))

                ## Optimizer
                prediction_df_era_new, correlations_by_era = score_numerai(
                    prediction_df_era,
                    features_raw_era,
                    features_optimizer,
                    proportion=proportion,
                    modelname=f"{modelname}-optimizer",
                    target_col_name=target_col[0],
                    era_col=era_col,
                    debug=debug,
                )
                prediction_dynamic.append(prediction_df_era_new.copy())
                correlation_dynamic.append(correlations_by_era)

                ### Dynamic Feature Neutralisation by different criteria
                DFN_params = list()
                for flavour in [
                    "mean",
                    "volatility",
                ]:
                    for direction in [
                        "tail",
                        "head",
                    ]:
                        for size in [
                            420,
                        ]:
                            if direction == "tail":
                                name = f"high_{flavour}_"
                            else:
                                name = f"low_{flavour}_"
                            if size == 420:
                                name = name + "standard"
                            elif size == 105:
                                name = name + "small"
                            temp = (flavour, size, direction, name)
                            DFN_params.append(temp)

                for DFN_param in DFN_params:
                    flavour = DFN_param[0]
                    factor_flavour_era = factor_flavour_eras[flavour][
                        factor_flavour_eras[flavour]["era"] == i
                    ]
                    selected_features = getattr(
                        factor_flavour_era.sort_values(flavour), DFN_param[2]
                    )(DFN_param[1])["feature_name"]

                    selected_features = list(set(selected_features) - set(bad_features))

                    prediction_df_era_new, correlations_by_era = score_numerai(
                        prediction_df_era,
                        features_raw_era,
                        selected_features,
                        proportion=proportion,
                        modelname=f"{modelname}-{DFN_param[3]}",
                        target_col_name=target_col[0],
                        era_col=era_col,
                        debug=debug,
                    )
                    prediction_dynamic.append(prediction_df_era_new.copy())
                    correlation_dynamic.append(correlations_by_era)

        return pd.concat(prediction_dynamic, axis=0), pd.concat(
            correlation_dynamic, axis=0
        )


def save_model_performance_test(
    Numerai_Model_Names,
    feature_corr,
    features_optimizer,
    startera=None,
    endera=None,
    data_file="data/v4_all_int8.parquet",
    data_version="v4",
    target_col=["target"],
    debug=False,
    gbm_start_iteration=0,
):
    (
        average_prediction_df,
        prediction_df_list,
    ) = predict_numerai_multiple(
        Numerai_Model_Names,
        feature_corr,
        filename=data_file,
        data_version=data_version,
        startera=startera,
        endera=endera,
        debug=debug,
        target_col=target_col,
        gbm_start_iteration=gbm_start_iteration,
    )

    del prediction_df_list

    MODEL_NAME = Numerai_Model_Names[0].split(".parameters")[0].split("/")[-1]
    MODEL_NAME = MODEL_NAME + f"_{len(Numerai_Model_Names)}"

    (
        features,
        targets,
        groups,
        weights,
    ) = load_numerai_data(
        data_file,
        resample_freq=1,
        startera=startera,
        endera=endera,
        target_col=target_col,
        data_version=data_version,
    )

    dynamic_predictions, dynamic_correlations = dynamic_feature_neutralisation(
        average_prediction_df,
        features,
        feature_corr,
        features_optimizer,
        target_col=target_col,
        modelname=MODEL_NAME,
        debug=debug,
    )
    summary_correlations = dynamic_correlations.pivot(
        index="era", columns="model_name", values=["neutralised_correlation"]
    ).dropna()
    strategy_flavour = pd.DataFrame.from_records(
        summary_correlations.apply(strategy_metrics, axis=0),
        index=summary_correlations.columns,
    )
    if data_version == "signals":
        return (
            strategy_flavour,
            summary_correlations,
            dynamic_predictions,
        )
    else:
        return (
            strategy_flavour,
            create_era_index(summary_correlations),
            dynamic_predictions,
        )


## Run Numerai Model Performances for both Classic and Signals tournament
def run_numerai_models_performances(
    Numerai_Model_Names,
    feature_corr,
    features_optimizer,
    PERFORMANCES_FOLDER,
    data_file="data/v4_all_int8.parquet",
    data_version="v4",
    target_col=["target"],
    gbm_start_iteration=0,
):
    ## Calculate Starting Era
    parametername = Numerai_Model_Names[0]
    no_models = len(Numerai_Model_Names)
    stem = parametername.split("/")[-1].replace(".parameters", "")
    correlations_filename = f"{PERFORMANCES_FOLDER}/{stem}_{no_models}.csv"
    if os.path.exists(parametername):
        parameters = joblib.load(parametername)
        if data_version == "signals":
            test_start = parameters["parameters"]["model"]["validation_end"]
            test_end = datetime.datetime.strptime("2099-12-31", "%Y-%m-%d")
        else:
            test_start = shift_era(
                parameters["parameters"]["model"]["validation_end"], gap=14
            )
            test_end = feature_corr.index[-1]
        if os.path.exists(correlations_filename):
            most_recent_date = pd.read_csv(correlations_filename, index_col=0).index[-1]
            if data_version == "signals":
                test_start = datetime.datetime.strptime(most_recent_date, "%Y-%m-%d")
            else:
                test_start = shift_era(convert_datetime_to_era(most_recent_date), gap=1)
        print(f"Model Performances {test_start} {test_end}")
        ### Get Model Predictions for the latest eras
        if test_end >= test_start:
            (
                validate_performance,
                validate_correlations,
                validate_predictions,
            ) = save_model_performance_test(
                Numerai_Model_Names,
                feature_corr,
                features_optimizer,
                startera=test_start,
                endera=test_end,
                data_file=data_file,
                data_version=data_version,
                target_col=target_col,
                gbm_start_iteration=gbm_start_iteration,
            )
            ## Update Model Performances
            output = validate_correlations["neutralised_correlation"]
            if os.path.exists(correlations_filename):
                old_file = pd.read_csv(correlations_filename, index_col=0)
                df = pd.concat([old_file, output.dropna()])
                df.index = pd.to_datetime(df.index)
                df[~df.index.duplicated()].sort_index().to_csv(correlations_filename)
            else:
                output.dropna().to_csv(correlations_filename)



"""
Numerai Signals 
"""


from joblib import Parallel, delayed
import pandas as pd
import numpy as np
from sklearn.base import TransformerMixin, BaseEstimator

import pycatch22

from .feature import SignatureTransformer


## Feature Transformation per era
def normalise_features_era(df, feature_cols, group_labels=None, keep_original=False):
    transformed_features = list()
    if group_labels is not None:
        for group in group_labels:
            group_features = list()
            for i, df_group in df.groupby(group):
                df_group_ranked = df_group[feature_cols].rank(pct=True, axis=0) - 0.5
                df_group_ranked.fillna(0, inplace=True)
                df_group_ranked = df_group_ranked * 5
                df_group_ranked.columns = [
                    "{}_{}_ranked".format(x, group) for x in feature_cols
                ]
                group_features.append(pd.concat([df_group_ranked], axis=1))
            group_features_df = pd.concat(group_features, axis=0)
            transformed_features.append(group_features_df)
    ## On All Data
    df_ranked = df[feature_cols].rank(pct=True, axis=0) - 0.5
    df_ranked.fillna(0, inplace=True)
    df_ranked = df_ranked * 5
    df_ranked.columns = ["{}_ranked".format(x) for x in feature_cols]
    transformed_features.append(df_ranked)
    if keep_original:
        transformed_features.append(df[feature_cols])
    transformed_df = pd.concat(transformed_features, axis=1)
    return transformed_df


class NumeraiSignatureTransformer(TransformerMixin, BaseEstimator):
    def __init__(self, signature_level=4, lookbacks=[4, 12, 52]):
        self.signature_level = signature_level
        self.lookbacks = lookbacks
        self.signaturetransformers = dict()
        for lookback in lookbacks:
            self.signaturetransformers[lookback] = SignatureTransformer(
                lookback, signature_level
            )

    def transform(self, X):
        signature_outputs = list()
        for lookback in self.lookbacks:
            signature_outputs.append(
                self.signaturetransformers[lookback].transform(X.dropna())
            )

        output = pd.concat(signature_outputs, axis=1)

        return output.dropna().add_prefix("feature_")


class NumeraiStatsTransformer(TransformerMixin, BaseEstimator):
    def __init__(
        self,
        lookbacks=[
            4,
            12,
            52,
        ],
    ):
        self.lookbacks = lookbacks

    def transform(self, X):
        stats_outputs = list()
        for lookback in self.lookbacks:
            for operation in ["mean", "std", "skew", "kurt"]:
                agg_X = getattr(X.rolling(lookback), operation)()
                agg_X.columns = [f"{col}_{operation}_{lookback}" for col in X.columns]
                stats_outputs.append(agg_X)
        output = pd.concat(stats_outputs, axis=1)

        return output.dropna().add_prefix("feature_")


def process_single_stock(
    df,
    shift=4,
    live_data=False,
    selected_cols=[
        "target_20d",
        "target_20d_raw_return",
        "target_20d_factor_neutral",
        "target_20d_factor_feat_neutral",
    ],
):
    shift_df = df[selected_cols].shift(shift) - 0.5
    if not live_data:
        ## Add Basic Stats Features
        transformer = NumeraiStatsTransformer()
        stats_features = transformer.transform(shift_df)
        features = stats_features
        features[selected_cols] = df[selected_cols]
        return features.dropna()
    else:
        MAX_LOOKBACK = 252
        transformer = NumeraiStatsTransformer()
        stats_features = transformer.transform(shift_df.tail(MAX_LOOKBACK))
        features = stats_features
        return features.tail(1)



from abc import ABC, abstractmethod
from typing import Optional, Sequence, Union, Type, Dict, List, Iterable

import pandas as pd
import numpy as np
import os.path

from .compustat import (
    merge_compustat_dataset,
    read_crsp_price,
    read_compustat_us_fundamentals,
    read_compustat_us_short_interests,
    read_optionmetrics_us_volatility_surface,
)
from .compustat import (
    read_compustat_price,
    read_compustat_international_fundamentals,
)


### Loading data downloaded from Compustat and CRSP


def Compustat_CRSP_Data(
    folder="../../data/CRSP",
    linktable="../../data/Compustat_CRSP_link_2021.csv",
    crsptable="../../data/CRSP_names_2021.csv",
    omtable="../../data/Option_Metrics_names_2021.csv",
    numeraitable="../Numerai-Signals/data/numerai_signals_metadata_2021.csv",
    fundamentals_folder="../../data/Compustat_Fundamentals",
    om_folder="../../data/Option_Metrics",
    market_index="Output/CRSP_USCore_all_all_500_index.csv",
    market="USCore",
    sectors=None,
    inds=None,
    subinds=None,
    startyear=1991,
    endyear=2021,
    quantile=5,
    use_fundamentals=False,
    use_option_volume=False,
    use_vol_surface=False,
    debug=False,
    identifier="permno",
):
    pricedf = read_crsp_price(
        folder=folder,
        linktable=linktable,
        crsptable=crsptable,
        omtable=omtable,
        numeraitable=numeraitable,
        market=market,
        sectors=sectors,
        inds=inds,
        subinds=subinds,
        startyear=startyear,
        endyear=endyear,
        debug=debug,
    )
    if pricedf.shape[0] > 0:
        has_price_data = True
    else:
        has_price_data = False

    ### Placeholder
    has_funda_data = False
    has_shortint_data = False
    has_options = False
    fundadf = None
    shortintdf = None
    optionsdf = None

    ## Always read short interests if possible

    selected_gvkeys = pricedf["gvkey"].dropna().unique()

    shortintdf = read_compustat_us_short_interests(
        fundamentals_folder=fundamentals_folder,
        startyear=startyear,
        endyear=endyear,
        selected_gvkeys=selected_gvkeys,
        debug=debug,
    )
    if shortintdf.shape[0] > 0:
        has_shortint_data = True

    if use_fundamentals:
        fundadf = read_compustat_us_fundamentals(
            fundamentals_folder=fundamentals_folder,
            startyear=startyear,
            endyear=endyear,
            selected_gvkeys=selected_gvkeys,
            debug=debug,
        )

        if fundadf.shape[0] > 0:
            has_funda_data = True

    if use_option_volume:
        selected_secids = pricedf["secid"].dropna().unique()

        optionsdf = read_optionmetrics_us_volatility_surface(
            om_folder=om_folder,
            startyear=startyear,
            endyear=endyear,
            use_vol_surface=use_vol_surface,
            selected_secids=selected_secids,
            debug=debug,
        )

        if optionsdf.shape[0] > 0:
            has_options = True

    if has_price_data:
        pivotdf = merge_compustat_dataset(
            pricedf,
            has_fundamentals=has_funda_data,
            fundadf=fundadf,
            has_short_interest=has_shortint_data,
            shortintdf=shortintdf,
            has_options=has_options,
            optionsdf=optionsdf,
            identifier=identifier,
            quantile=quantile,
            debug=debug,
        )
        if pivotdf.shape[0] > 0:
            return pivotdf
        else:
            return pd.DataFrame()
    else:
        return pd.DataFrame()


def Compustat_Data(
    folder="../../data/Compustat",
    fundamentals_folder="../../data/Compustat_Fundamentals",
    usnameref="../../data/Compustat_metadata_2020.csv",
    market="US",
    startyear=2016,
    endyear=2021,
    sectors=None,
    inds=None,
    subinds=None,
    quantile=5,
    debug=False,
):
    pricedf = read_compustat_price(
        folder=folder,
        market=market,
        startyear=startyear,
        endyear=endyear,
        sectors=sectors,
        inds=inds,
        subinds=subinds,
        debug=debug,
    )

    ## Compute percentage return
    if pricedf.shape[0] > 0:
        has_pricedata = True
    else:
        has_pricedata = False

    if market == "US":
        fundadf = read_compustat_us_fundamentals(
            fundamentals_folder=fundamentals_folder,
            startyear=startyear,
            endyear=endyear,
            debug=debug,
        )
        ## Use cusip not gvkey as identifier as gvkey is NOT unique for companies with multiple share classes
        identifier = "cusip"
    else:
        ## Not Implemented yet
        # fundadf = read_compustat_international_fundamentals(fundamentals_folder=fundamentals_folder, startyear=startyear, endyear=endyear,)
        fundadf = pd.DataFrame()
        ## Must use isin as identifier since gvkey is NOT unique for companies with multiple share classes
        identifier = "isin"

    if fundadf.shape[0] > 0:
        has_funda_data = True
    else:
        has_funda_data = False

    ### To be Implemented
    ### US Compustat Short Interest
    has_shortint_data = False
    shortintdf = pd.DataFrame()

    if has_pricedata:
        ## Merge
        pivotdf = merge_compustat_dataset(
            pricedf,
            has_funda_data,
            fundadf,
            has_shortint_data,
            shortintdf,
            identifier=identifier,
            quantile=quantile,
            debug=debug,
        )
        if pivotdf.shape[0] > 0:
            return pivotdf
        else:
            return pd.DataFrame()
    else:
        return pd.DataFrame()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Process Data from Compustat Database
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


# ##### Option Metrics
# # Merging between CRSP and Option Metrics are based on cusip
# ## key field in Option Metrics are secid, key field in CRSP are permno
# ## Use cusip(8-digit) in Option Metrics to link with ncusip in CRSP (which reflects point in time)
# ###


import pandas as pd
import numpy as np
import os.path
import datetime

#### Read a single year price data from CRSP
def read_crsp_price_data(folder, year):
    ## Mapping CRSP columns
    column_map = {
        "ret": "return",
        "retx": "price_return",
        "shrout": "share_outstanding",
        "openprc": "open",
        "bidlo": "low",
        "askhi": "high",
        "prc": "close",
        "vol": "volume",
        "cfacpr": "price_adjust_factor",
        "cfacshr": "share_adjust_factor",
        "numtrd": "trades",
        "date": "date",
        "cusip": "cusip8",
        "permno": "permno",
        "permco": "permco",
    }

    ## Read One Year of Price Data
    pricefile = "{}/CRSP_{}_price.parquet".format(folder, year)
    if os.path.isfile(pricefile):
        pricedf = pd.read_parquet(pricefile)
    else:
        return None

    pricedf.rename(columns=column_map, inplace=True)
    ## Set Data Types
    pricedf["date"] = pd.to_datetime(pricedf["date"])
    ## Compute Dollar Volume, Market Cap
    pricedf["market_cap"] = pricedf["close"].abs() * pricedf["share_outstanding"] * 1000
    pricedf["dollar_volume"] = pricedf["close"].abs() * pricedf["volume"]
    pricedf["spread"] = np.abs(pricedf["ask"] - pricedf["bid"]) / pricedf["close"].abs()
    ## Split adjust the prices, do not adjust for dividends to avoid looking ahead
    for col in ["close", "open", "high", "low"]:
        pricedf["adjusted_{}".format(col)] = (
            pricedf[col].abs() / pricedf["share_adjust_factor"]
        )

    num_cols = [
        "market_cap",
        "dollar_volume",
        "adjusted_close",
        "adjusted_open",
        "adjusted_high",
        "adjusted_low",
    ]
    for num_col in num_cols:
        pricedf[num_col] = pricedf[num_col].astype(float)

    ### Pivot
    pricedf.dropna(
        subset=["permno", "date", "dollar_volume", "market_cap"], axis=0, inplace=True
    )
    return pricedf


###  Read CRSP Price Data
###  Merge with Compustat CRSP Link to obtain link to Compustat
###  Merge with CRSP Names Table to obtain share class and start date
###  Merge with Option Metrics Table to obtain link to OM
###  Merge with Numerai Table to obtain link to bloomberg ticker and universe filter
###
def read_crsp_price(
    basefolder="../../data",
    crspfolder="CRSP",
    linktable="Compustat_CRSP_link_2021.csv",
    crsptable="CRSP_names_2021.csv",
    omtable="Option_Metrics_names_2021.csv",
    numeraitable="numerai_signals_metadata_2021.csv",
    market="USCore",
    sectors=None,
    inds=None,
    subinds=None,
    startyear=1921,
    endyear=2021,
    debug=False,
):
    def universe_filter(
        pricedf,
        market="USCore",
        sectors=None,
        inds=None,
        subinds=None,
    ):
        if market == "Numerai":
            numerai_tickers = pd.read_csv(f"{numeraitable}").drop_duplicates(
                subset=["permno", "bloomberg_ticker"]
            )
            numerai_tickers["sector"] = numerai_tickers["hgsubind"] // 1000000
            numerai_tickers["industry"] = numerai_tickers["hgsubind"] // 100
            if sectors:
                numerai_tickers = numerai_tickers[
                    numerai_tickers["sector"].isin(sectors)
                ]
            if inds:
                numerai_tickers = numerai_tickers[
                    numerai_tickers["industry"].isin(inds)
                ]
            pricedf = pricedf[pricedf["permno"].isin(numerai_tickers["permno"])]

        if market == "Commodities":
            commodities = [
                ## Commodities Index
                91129,  ## DBC
                91712,  ## DBA Agriculture
                91715,  ## DBB
                91710,  ## DBP
                91710,  ## DBE
                ## US Bonds
                89468,  ## TLT
                91933,  ## HYG
                89467,  ## LQD
                ## Dollar Index
                91758,  ## UUP
                ## US Market
                85765,  ## DIA
                84398,  ## SPY
                86755,  ## QQQ
                88222,  ## IWM
                ## Stock Sector
                86449,  ## XLB
                86451,  ## XLV
                86452,  ## XLP
                86453,  ## XLY
                86454,  ## XLE
                86455,  ## XLF
                86456,  ## XLI
                86457,  ## XLK
                86458,  ## XLU
                17940,  ## XLC
                15732,  ## XLRE
                ## Commodities Producers
                91232,  ## GDX
                93318,  ## SIL
                93319,  ## COPX
                12372,  ## URA
            ]

            pricedf = pricedf[pricedf["permno"].isin(commodities)]

        return pricedf.copy()

    pricedfs = list()

    output_columns = [
        "cusip",
        "date",
        "close",
        "adjusted_close",
        "adjusted_high",
        "adjusted_open",
        "adjusted_low",
        "spread",
        "return",
        "dollar_volume",
        "market_cap",
        "permno",  ## CRSP Primary Key
        "gvkey",  ## Compustat Foreign Key
        "secid",  ## Option Metrics Foregin Key
        # "cik",  ## SEC EDGAR Foreign Key
        # "tic",  ## Numerai Foregin Key
        # "gind",
        # "gsubind",
        # "sector",
        # "conm",
    ]

    for year in range(startyear, endyear + 1):
        ### Filtering is based on previous year average market cap greater than threshold
        pricedf = read_crsp_price_data(f"{basefolder}/{crspfolder}", year)

        if pricedf is not None:
            ## Merge with Compustat Link Table
            if os.path.isfile(f"{basefolder}/{linktable}"):
                linktabledf = pd.read_csv(f"{basefolder}/{linktable}").dropna(
                    subset=["gvkey", "lpermno", "lpermco"]
                )
                linktabledf.drop_duplicates(
                    subset=[
                        "gvkey",
                        "lpermno",
                        "lpermco",
                    ],
                    inplace=True,
                )
                pricedf = pricedf.merge(
                    linktabledf[
                        [
                            "lpermno",
                            "lpermco",
                            "cik",
                            "gind",
                            "gsubind",
                            "cusip",
                            "conm",
                            "tic",
                            "gvkey",
                        ]
                    ],
                    left_on=["permno", "permco"],
                    right_on=["lpermno", "lpermco"],
                    how="left",
                )

                pricedf.drop(columns=["lpermno", "lpermco"], axis=1, inplace=True)

            ## Merge with Option Metrics Table
            if os.path.isfile(f"{basefolder}/{omtable}"):
                om = (
                    pd.read_csv(f"{basefolder}/{omtable}")
                    .dropna(subset=["ticker"])
                    .drop_duplicates(subset=["cusip", "secid"])
                )
                om["omcusip"] = om["cusip"].copy()
                pricedf = pricedf.merge(
                    om[["omcusip", "secid"]],
                    left_on="cusip8",
                    right_on="omcusip",
                    how="left",
                )

            ## Remove entries without price and trading data
            pricedf.dropna(
                subset=[
                    "market_cap",
                    "dollar_volume",
                    "permno",
                    "cusip",
                ],
                inplace=True,
            )

            pricedf.drop_duplicates(subset=["permno", "cusip", "date"], inplace=True)

            ## Filter down to asset within the required industries and market cap
            pricedf = universe_filter(pricedf, market, sectors, inds, subinds)

            if debug:
                print("CRSP Price ", year, pricedf.shape)

            pricedfs.append(pricedf[output_columns])

    if len(pricedfs) > 0:
        return pd.concat(pricedfs, axis=0, ignore_index=True)
    else:
        return pd.DataFrame()


#### Read Compustat US Fundmentals Data
def read_compustat_us_fundamentals(
    fundamentals_folder="../../data/Compustat_Fundamentals",
    startyear=2016,
    endyear=2021,
    selected_gvkeys=None,
    debug=False,
):
    funda_years = list()

    for datayear in range(startyear - 1, endyear + 1, 1):
        fundafile = "{}/us_fundamentals_quarter_{}.parquet".format(
            fundamentals_folder, datayear
        )
        if os.path.isfile(fundafile):
            funda_year = pd.read_parquet(fundafile)
            funda_year["gvkey"] = funda_year["gvkey"].astype(int)
            funda_year = funda_year[funda_year["gvkey"].isin(selected_gvkeys)]
            funda_year.dropna(
                subset=[
                    "rdq",
                ],
                inplace=True,
            )
            funda_year.sort_values(["gvkey", "rdq"], inplace=True)
            funda_year.drop_duplicates(["gvkey", "rdq"], inplace=True)
            if debug:
                print("Compustat US Fundamentals ", datayear, funda_year.shape)
            funda_years.append(funda_year)

    if len(funda_years) > 0:
        funda_sample = pd.concat(funda_years, axis=0)
        ## Calculating financial data
        funda_sample["buyback"] = funda_sample["cshopq"] * funda_sample["prcraq"]
        funda_sample["dividend"] = funda_sample["dvy"]

        funda_sample["earnings"] = funda_sample["niq"]
        funda_sample["sales"] = funda_sample["saleq"]
        funda_sample["cogs"] = funda_sample["cogsq"]  ## cost of goods sold
        funda_sample["inventory"] = funda_sample["invtq"]
        funda_sample["receivables"] = funda_sample["rectq"]
        funda_sample["income_before_extra_items"] = funda_sample["ibq"]
        funda_sample["pretax_income"] = funda_sample["piq"]
        funda_sample["interest_expense"] = funda_sample["xintq"]

        funda_sample["operating_cashflow"] = funda_sample["oancfy"]
        funda_sample["capital_expenditure"] = funda_sample["capxy"]

        funda_sample["book_value"] = funda_sample["ceqq"]  ## common equity
        funda_sample["intangible_book_value"] = (
            funda_sample["intanq"] - funda_sample["ltq"]
        )
        funda_sample["total_asset"] = funda_sample["atq"]
        funda_sample["total_liability"] = funda_sample["ltq"]
        funda_sample["total_equity"] = funda_sample["teqq"]
        funda_sample["intangible_asset"] = funda_sample["intanq"]
        funda_sample["invested_capital"] = funda_sample["icaptq"]
        funda_sample["longterm_debt"] = funda_sample["dlttq"]
        funda_sample["current_asset"] = funda_sample["actq"]
        funda_sample["current_liability"] = funda_sample["lctq"]
        funda_sample["cash"] = funda_sample["cheq"]  ## cash and short term investment

        ## Financial Report Date
        funda_sample["report_date"] = pd.to_datetime(funda_sample["rdq"])

        ## Fill missing financials with zeros
        financial_cols = [
            "buyback",
            "dividend",
            "earnings",
            "sales",
            "inventory",
            "cogs",
            "receivables",
            "income_before_extra_items",
            "pretax_income",
            "interest_expense",
            "operating_cashflow",
            "capital_expenditure",
            "book_value",
            "intangible_book_value",
            "total_asset",
            "total_liability",
            "total_equity",
            "invested_capital",
            "longterm_debt",
            "current_asset",
            "current_liability",
            "cash",
        ]
        for col in financial_cols:
            funda_sample[col] = funda_sample[col].fillna(0)
        ## Process financial data,
        output_cols = financial_cols + [
            "report_date",
            "gvkey",
        ]
        return funda_sample[output_cols].copy()
    else:
        return pd.DataFrame()


def read_compustat_us_short_interests(
    fundamentals_folder="../../data/Compustat_Fundamentals",
    startyear=2016,
    endyear=2021,
    selected_gvkeys=None,
    debug=False,
):
    shortinterests = list()
    for year in range(startyear, endyear + 1):
        filename = "{}/us_short_interest_{}.parquet".format(
            fundamentals_folder,
            year,
        )
        if os.path.isfile(filename):
            sample = pd.read_parquet(filename)
            sample = sample[sample["gvkey"].isin(selected_gvkeys)]
            sample["date"] = pd.to_datetime(sample["datadate"])
            if debug:
                print("Compustat US Short Int ", year, sample.shape)
            shortinterests.append(
                sample[
                    [
                        "date",
                        "gvkey",
                        "shortint",
                    ]
                ]
            )

    if len(shortinterests) > 0:
        return pd.concat(shortinterests, axis=0)
    else:
        return pd.DataFrame()


# ### Option Metrics Data
# ##
# ##
# ##


def read_optionmetrics_us_volatility_surface(
    om_folder="../../data/Option_Metrics",
    startyear=2016,
    endyear=2021,
    use_vol_surface=False,
    selected_secids=None,
    debug=True,
):
    if use_vol_surface:
        volsurfaces = list()
        for year in range(startyear, endyear + 1):
            for month in range(1, 13):
                filename = "{}/volatility_surface_{}_{}.parquet".format(
                    om_folder, year, month
                )
                if os.path.isfile(filename):
                    sample = pd.read_parquet(filename)
                    ### Filter to include only volatility surface within a year to reduce the number of columns
                    sample = sample[
                        (sample["days"] >= 30)
                        & (sample["days"] <= 252)
                        & (sample["delta"] <= 60)
                        & (sample["delta"] >= -60)
                    ]
                    ## Filter to include data for selected stocks only to reduce memory usage
                    sample = sample[sample["secid"].isin(selected_secids)]
                    ### Make column names
                    sample["date"] = pd.to_datetime(sample["date"])
                    sample["days"] = sample["days"].apply(str)
                    sample["delta"] = sample["delta"].apply(str)
                    volsurface = sample.pivot(
                        index=[
                            "secid",
                            "date",
                        ],
                        columns=["days", "delta"],
                        values="impl_volatility",
                    )
                    volsurface.columns = [
                        "IV_" + "_".join(column).rstrip("_")
                        for column in volsurface.columns.to_flat_index()
                    ]
                    if debug:
                        print(
                            "Option Metrics Volatility Surface ",
                            year,
                            month,
                            volsurface.shape,
                        )
                    volsurfaces.append(volsurface)
        volsurfaces_merged = pd.concat(volsurfaces, axis=0)
    ## Option Volumes
    optionvolumes = list()
    for year in range(startyear, endyear + 1):
        filename = "{}/option_volume_{}.parquet".format(om_folder, year)
        if os.path.isfile(filename):
            sample = pd.read_parquet(filename)
            sample = sample[sample["secid"].isin(selected_secids)]
            sample["date"] = pd.to_datetime(sample["date"])
            sample["cp_flag"].fillna("A", inplace=True)
            optionvolume = sample.pivot(
                index=["secid", "date"],
                columns=["cp_flag"],
                values=["volume", "open_interest"],
            )
            optionvolume.columns = [
                "option_" + "_".join(column).rstrip("_")
                for column in optionvolume.columns.to_flat_index()
            ]
            if debug:
                print("Option Metrics Option Volume ", year, optionvolume.shape)
            optionvolumes.append(optionvolume)
    optionvolumes_merged = pd.concat(optionvolumes, axis=0)
    if use_vol_surface:
        return pd.concat(
            [volsurfaces_merged, optionvolumes_merged], axis=1
        ).reset_index()
    else:
        return optionvolumes_merged.reset_index()


###
### Given price data from CRSP or Compustat, Merge with Fundamentals data from Compustat, Short Interest from Compustat
###
def merge_compustat_dataset(
    pricedf,
    has_fundamentals=False,
    fundadf=None,
    has_short_interest=False,
    shortintdf=None,
    has_options=False,
    optionsdf=None,
    identifier="cusip",
    quantile=5,
    debug=False,
):
    price_cols = list(pricedf.columns)
    selected_cols = list(pricedf.columns)

    if has_fundamentals:
        pricedf = pricedf.merge(
            fundadf,
            left_on=["gvkey", "date"],
            right_on=["gvkey", "report_date"],
            how="left",
        )
        selected_cols.extend(fundadf.columns)
        if debug:
            print("With Fundamentals", pricedf.shape)

    if has_short_interest:
        pricedf = pricedf.merge(
            shortintdf,
            left_on=["gvkey", "date"],
            right_on=["gvkey", "date"],
            how="left",
        )
        selected_cols.extend(shortintdf.columns)
        if debug:
            print("All Short Int ", pricedf.shape)

    ### Add additional Compustat Datasets Here
    ### Compustat Institutioal Holders
    ### Compustat Insiders

    ### Add Option Metrics Data Here
    if has_options:
        pricedf = pricedf.merge(
            optionsdf,
            left_on=["secid", "date"],
            right_on=["secid", "date"],
            how="left",
        )
        selected_cols.extend(optionsdf.columns)
        if debug:
            print("All Options ", pricedf.shape)

    ### pivot dataframe
    selected_cols = list(set(selected_cols) - set([identifier, "date"]))
    pricedf.drop_duplicates(subset=[identifier, "date"], inplace=True)
    ### Filter universe so that we only consider stocks that are shortable and has options
    if has_short_interest:
        pricedf.dropna(
            axis=0,
            subset=[
                identifier,
                "gvkey",
            ],
            inplace=True,
        )
        if debug:
            print("Data with short interest ", pricedf.shape)

    if has_options:
        pricedf.dropna(
            axis=0,
            subset=[
                identifier,
                "secid",
            ],
            inplace=True,
        )
        if debug:
            print("Data with option volumes ", pricedf.shape)

    ### Pivot table into dates index and stocks,features as columns
    pivotdf = pricedf.pivot(index="date", columns=identifier, values=selected_cols)

    ### Create return column for Compustat International price data
    if not "return" in selected_cols and "adjusted_close" in selected_cols:
        ## Forward Fill five days as JPN has missing data in 2003
        returns = pivotdf[["adjusted_close"]].pct_change(fill_method=None)
        returns.rename(columns={"adjusted_close": "return"}, inplace=True)
        pivotdf = pd.concat([pivotdf, returns], axis=1)

    if debug:
        print("Pivoted Data", pivotdf.shape)

    if pivotdf.shape[0] > 0:
        ## Calculate Seasonality Features
        pivotdf["date"] = pivotdf.index.copy()
        pivotdf["year"] = pivotdf["date"].dt.year

        ### Trading days since start of year
        ### Use: Align rebalance dates across years
        ##
        pivotdf["trading_days"] = 0
        prev_year = pivotdf["year"].iloc[0]
        counter = -1
        for i, row in pivotdf[["date", "year"]].iterrows():
            current_year = row["year"][0]
            if current_year == prev_year:
                counter += 1
            else:
                counter = 0
            pivotdf.loc[i, "trading_days"] = counter
            prev_year = current_year

    return pivotdf


### Read Compustat Price Data and crate market cap based filters
###
###
def read_compustat_price_year(folder, market, year):
    ## Read price data from Compustat
    column_map = {
        "datadate": "date",
        "trfd": "total_return_factor",
        "ajexdi": "adjust_ratio",
        "cshoc": "share_outstanding",
        "cshtrd": "volume",
        "prccd": "close",
        "prcld": "open",
        "prchd": "high",
        "prcod": "low",
        "curcdd": "currency",
        "exchg": "exchange",
        "cusip": "cusip",
        "isin": "isin",
        "gsubind": "gsubind",
        "gind": "gind",
    }

    ### Read Previous Year Pircedata and create investable universe for next year based on market cap

    if market == "US":
        pricefilename = "{}_{}/{}_price_{}.parquet".format(folder, market, market, year)
        identifier = "cusip"
    else:
        pricefilename = "{}_International/{}_price_{}.parquet".format(
            folder, market, year
        )
        identifier = "isin"
    if os.path.isfile(pricefilename):
        pricedf = pd.read_parquet(pricefilename)
        if identifier in pricedf.columns:
            pricedf.set_index(["isin", "datadate"], inplace=True)
        pricedf.reset_index(inplace=True)
    else:
        return None, None

    pricedf.rename(columns=column_map, inplace=True)
    pricedf["date"] = pd.to_datetime(pricedf["date"])
    ## Compute Dollar Volume, Market Cap, Adjusted Price
    pricedf["market_cap"] = pricedf["close"] * pricedf["share_outstanding"]
    pricedf["dollar_volume"] = pricedf["close"] * pricedf["volume"]
    for col in ["close", "open", "high", "low"]:
        pricedf["adjusted_{}".format(col)] = (
            pricedf[col] / pricedf["adjust_ratio"] * pricedf["total_return_factor"]
        )

    pricedf.sort_values(
        [
            "date",
            "gvkey",
            "adjusted_close",
        ],
        inplace=True,
    )
    pricedf.drop_duplicates(subset=["date", "gvkey"], inplace=True)
    pricedf.dropna(
        subset=[
            "date",
            "gvkey",
            "market_cap",
            "dollar_volume",
        ],
        inplace=True,
        axis=0,
    )

    return pricedf


# ## Loading data downloaded from Compustat and CRSP


def Compustat_CRSP_Data(
    basefolder="../../data",
    crspfolder="CRSP",
    linktable="Compustat_CRSP_link_2021.csv",
    crsptable="CRSP_names_2021.csv",
    omtable="Option_Metrics_names_2021.csv",
    numeraitable="numerai_signals_metadata_2021.csv",
    fundamentals_folder="Compustat_Fundamentals",
    om_folder="Option_Metrics",
    market="USCore",
    sectors=None,
    inds=None,
    subinds=None,
    startyear=1991,
    endyear=2021,
    quantile=5,
    use_fundamentals=False,
    use_option_volume=False,
    use_vol_surface=False,
    debug=False,
    identifier="permno",
):
    pricedf = read_crsp_price(
        basefolder=basefolder,
        linktable=linktable,
        crsptable=crsptable,
        omtable=omtable,
        numeraitable=numeraitable,
        market=market,
        sectors=sectors,
        inds=inds,
        subinds=subinds,
        startyear=startyear,
        endyear=endyear,
        debug=debug,
    )
    if pricedf.shape[0] > 0:
        has_price_data = True
    else:
        has_price_data = False

    ### Placeholder
    has_funda_data = False
    has_shortint_data = False
    has_options = False
    fundadf = None
    shortintdf = None
    optionsdf = None

    ## Always read short interests if possible

    selected_gvkeys = pricedf["gvkey"].dropna().unique()

    shortintdf = read_compustat_us_short_interests(
        fundamentals_folder=f"{basefolder}/{fundamentals_folder}",
        startyear=startyear,
        endyear=endyear,
        selected_gvkeys=selected_gvkeys,
        debug=debug,
    )
    if shortintdf.shape[0] > 0:
        has_shortint_data = True

    if use_fundamentals:
        fundadf = read_compustat_us_fundamentals(
            fundamentals_folder=f"{basefolder}/{fundamentals_folder}",
            startyear=startyear,
            endyear=endyear,
            selected_gvkeys=selected_gvkeys,
            debug=debug,
        )

        if fundadf.shape[0] > 0:
            has_funda_data = True

    if use_option_volume:
        selected_secids = pricedf["secid"].dropna().unique()

        optionsdf = read_optionmetrics_us_volatility_surface(
            om_folder=f"{basefolder}/{om_folder}",
            startyear=startyear,
            endyear=endyear,
            use_vol_surface=use_vol_surface,
            selected_secids=selected_secids,
            debug=debug,
        )

        if optionsdf.shape[0] > 0:
            has_options = True

    if has_price_data:
        pivotdf = merge_compustat_dataset(
            pricedf,
            has_fundamentals=has_funda_data,
            fundadf=fundadf,
            has_short_interest=has_shortint_data,
            shortintdf=shortintdf,
            has_options=has_options,
            optionsdf=optionsdf,
            identifier=identifier,
            quantile=quantile,
            debug=debug,
        )
        if pivotdf.shape[0] > 0:
            return pivotdf
        else:
            return pd.DataFrame()
    else:
        return pd.DataFrame()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# A collection of GBDT models for temporal tabular data
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pandas as pd
import numpy as np
import joblib, os, shutil, datetime



import logging, gc

if torch.cuda.is_available():
    import cupy as cp


from sklearn.metrics import mean_squared_error
from sklearn.model_selection import KFold, GroupKFold, GroupShuffleSplit

## Machine Learning packages
from xgboost import XGBRegressor
from lightgbm import LGBMRegressor, LGBMClassifier
from catboost import CatBoostRegressor

import lightgbm, xgboost, catboost
import torch



from .util import align_features_target, RollingTSTransformer, GroupedTimeSeriesSplit
from .feature import benchmark_features_transform
from .neural import TabularModel, MLP, LSTM_Tabular

# ## Persistence of ML models


### Save Best Model using method provided
def save_best_model(model, model_type, outputpath):
    if model_type in [
        "lightgbm",
        "lightgbm-gbdt",
        "lightgbm-goss",
        "lightgbm-rf",
        "lightgbm-dart",
    ]:
        model.save_model(outputpath)
    if model_type in [
        "xgboost",
        "xgboost-dart",
        "xgboost-gbtree",
    ]:
        model.save_model(outputpath)
    if model_type == "catboost":
        model.save_model(outputpath)
    if model_type in [
        "Numerai-MLP",
        "Numerai-LSTM",
    ]:
        model.save_model(outputpath)
    if model_type == "tabnet":
        model.save_model(outputpath)
        os.rename("{}.zip".format(outputpath), outputpath)
    return None


### load Best Model using method provided
def load_best_model(model_type, outputpath):
    if model_type in [
        "lightgbm",
        "lightgbm-gbdt",
        "lightgbm-goss",
        "lightgbm-rf",
        "lightgbm-dart",
    ]:
        reg = lightgbm.Booster(model_file=outputpath)
    if model_type in [
        "xgboost",
        "xgboost-dart",
        "xgboost-gbtree",
    ]:
        reg = xgboost.Booster()
        reg.load_model(outputpath)
    if model_type == "catboost":
        reg = catboost.CatBoost()
        reg.load_model(outputpath)
    if model_type in [
        "Numerai-MLP",
    ]:
        reg = TabularModel(MLP, config=dict())
        reg.load_model(outputpath)
    if model_type in [
        "Numerai-LSTM",
    ]:
        reg = TabularModel(LSTM_Tabular, config=dict())
        reg.load_model(outputpath)
    if model_type == "tabnet":
        from pytorch_tabnet.tab_model import TabNetRegressor

        reg = TabNetRegressor()
        reg.load_model(outputpath)
    if model_type == "feature-momentum":
        reg = None
    return reg


# ## Fit ML Models


def benchmark_neural_model(
    extracted_features_train,
    y_train,
    weights_train,
    extracted_features_test=None,
    y_test=None,
    weights_test=None,
    tabular_model="Numerai-MLP",
    tabular_hyper=None,
    additional_hyper=None,
    debug=False,
):
    gc.collect()

    ## Initialise and Train Models
    if tabular_model in [
        "Numerai-MLP",
    ]:
        reg = TabularModel(MLP, config=tabular_hyper)
        reg.train(extracted_features_train, y_train, extracted_features_test, y_test)
        pred = reg.predict(extracted_features_test.values)
        return reg, pred

    if tabular_model in [
        "Numerai-LSTM",
    ]:
        reg = TabularModel(LSTM_Tabular, config=tabular_hyper)
        reg.train(extracted_features_train, y_train, extracted_features_test, y_test)
        pred = reg.predict(extracted_features_test.values)
        return reg, pred

    if tabular_model == "tabnet":
        ## Default is PyTorch Adam Optimizer
        from torch.optim import Adam
        from torch.optim.lr_scheduler import StepLR
        from pytorch_tabnet.tab_model import TabNetRegressor

        tabnet_hyper = dict()
        tabnet_hyper["optimizer_fn"] = Adam
        tabnet_hyper["optimizer_params"] = {
            "lr": 0.02,
        }
        tabnet_hyper["scheduler_fn"] = StepLR
        tabnet_hyper["scheduler_params"] = {"gamma": 0.95, "step_size": 20}

        for key in [
            "seed",
            "n_d",
            "n_a",
            "n_steps",
            "n_independent",
            "n_shared",
            "gamma",
            "momentum",
            "lambda_sparse",
        ]:
            tabnet_hyper[key] = tabular_hyper[key]

        ## Separate Hyper-parameters in the fit function
        tabnet_fit_hyper = dict()
        for key in [
            "max_epochs",
            "patience",
            "batch_size",
        ]:
            tabnet_fit_hyper[key] = tabular_hyper[key]

        reg = TabNetRegressor(**tabnet_hyper)
        reg.fit(
            extracted_features_train.values,
            y_train.values,
            eval_set=[(extracted_features_test.values, y_test.values)],
            max_epochs=tabnet_fit_hyper.get("max_epochs", 20),
            patience=tabnet_fit_hyper.get("patience", 5),
            batch_size=tabnet_fit_hyper.get("batch_size", 40960),
            virtual_batch_size=int(tabnet_fit_hyper.get("batch_size", 40960) / 4),
            num_workers=0,
        )
        pred = reg.predict(extracted_features_test.values)
        return reg, pred


def benchmark_tree_model(
    extracted_features_train,
    y_train,
    weights_train,
    extracted_features_test=None,
    y_test=None,
    weights_test=None,
    tabular_model="lightgbm",
    tabular_hyper=None,
    additional_hyper=None,
    debug=False,
):
    ### Free up Memory from previous loop
    gc.collect()

    #### Fit Regressor Model for different ML methods
    if tabular_model in [
        "lightgbm",
        "lightgbm-gbdt",
        "lightgbm-dart",
        "lightgbm-goss",
        "lightgbm-rf",
    ]:
        if y_test is not None:
            train_data = lightgbm.Dataset(
                extracted_features_train,
                label=y_train,
                weight=weights_train,
                params={"max_bin": tabular_hyper["max_bin"]},
            )
            test_data = lightgbm.Dataset(
                extracted_features_test,
                label=y_test,
                weight=weights_test,
                params={"max_bin": tabular_hyper["max_bin"]},
            )
            early_stopping_rounds = tabular_hyper.get("early_stopping_round", 0)
            model = lightgbm.train(
                tabular_hyper,
                train_set=train_data,
                num_boost_round=tabular_hyper["num_iterations"],
                valid_sets=[test_data],
                callbacks=[
                    lightgbm.log_evaluation(period=1000),
                    lightgbm.early_stopping(early_stopping_rounds),
                ],
            )
            valid_iteration = min(
                additional_hyper.get("gbm_start_iteration", 0),
                int(model.num_trees() // 2),
            )
            pred = model.predict(
                extracted_features_test, start_iteration=valid_iteration
            )
            return model, pred
        else:
            train_data = lightgbm.Dataset(
                extracted_features_train,
                label=y_train,
                weight=weights_train,
            )
            model = lightgbm.train(
                tabular_hyper,
                train_set=train_data,
                num_boost_round=tabular_hyper["num_iterations"],
            )
            return model

    ## xgboost ignores extra parameters
    if tabular_model in [
        "xgboost",
        "xgboost-dart",
        "xgboost-gbtree",
    ]:
        ## Create DMatrix
        if y_test is not None:
            train_data = xgboost.DMatrix(
                extracted_features_train,
                label=y_train.values.reshape(-1),
                weight=weights_train.values.reshape(-1),
            )
            test_data = xgboost.DMatrix(
                extracted_features_test,
                label=y_test.values.reshape(-1),
                weight=weights_test.values.reshape(-1),
            )
            ### Train XGBoost model
            model = xgboost.train(
                tabular_hyper,
                train_data,
                num_boost_round=tabular_hyper["num_boost_round"],
                evals=[(test_data, "xgboost_test_data")],
                early_stopping_rounds=tabular_hyper["early_stopping_rounds"],
                verbose_eval=100,
            )
            start_iteration = min(
                additional_hyper.get("gbm_start_iteration", 0),
                int(model.best_iteration // 2),
            )
            end_iteration = model.best_iteration
            pred = model.predict(
                test_data,
                iteration_range=(start_iteration, end_iteration),
            )
            return model, pred
        else:
            train_data = xgboost.DMatrix(
                extracted_features_train,
                label=y_train.values.reshape(-1),
                weight=weights_train.values.reshape(-1),
            )
            model = xgboost.train(
                tabular_hyper,
                train_data,
                num_boost_round=tabular_hyper["num_boost_round"],
            )
            return model


### Run ML pipeline for temporal tabular data
def benchmark_pipeline(
    features,
    target,
    weights,
    groups,
    model_params=None,
    feature_eng=None,
    feature_eng_parameters=None,
    tabular_model="lightgbm",
    tabular_hyper=None,
    additional_hyper=None,
    debug=False,
):
    if debug:
        print(f"Dataset Sizes {features.shape} {target.shape} {groups.shape}")

    if not model_params:
        model_params = {
            "valid_splits": 1,
            "test_size": 52,
            "max_train_size": 52,
            "gap": 52,
            "cross_validation": "GroupedTimeSeriesSplit",
        }

    ## Cross Validation split
    if model_params["cross_validation"] == "GroupedTimeSeriesSplit":
        tscv = GroupedTimeSeriesSplit(
            valid_splits=model_params["valid_splits"],
            test_size=model_params["test_size"],
            max_train_size=model_params["max_train_size"],
            gap=model_params["gap"],
            debug=debug,
        )
    elif model_params["cross_validation"] == "GroupShuffleSplit":
        tscv = GroupShuffleSplit(
            n_splits=model_params["n_splits"],
            test_size=model_params["test_size"],
            train_size=model_params["train_size"],
            random_state=model_params.get("random_state", 0),
        )
    else:
        tscv = KFold(
            n_splits=model_params["n_splits"],
            shuffle=True,
            random_state=model_params.get("random_state", 0),
        )
    model_no = 1
    model_performance = dict()
    trained_models = dict()
    data = dict()
    parameters = dict()

    for train_index, test_index in tscv.split(features, groups=groups):
        ## Get Trained and Test Data
        if model_params["cross_validation"] == "GroupedTimeSeriesSplit":
            X_train, X_test = features.loc[train_index, :], features.loc[test_index, :]
            y_train, y_test = target.loc[train_index, :], target.loc[test_index, :]
            ## Data Weights are pd Series
            weights_train, weights_test = (
                weights.loc[train_index],
                weights.loc[test_index],
            )
            ## Group Labels are pd Series
            group_train, group_test = (
                groups.loc[train_index],
                groups.loc[test_index],
            )
            if debug:
                print(X_train.shape, X_test.shape)

        ## For Existing Cross Validation Splits in scikit-learn it is based on index location (iloc)
        else:
            X_train, X_test = (
                features.iloc[train_index, :],
                features.iloc[test_index, :],
            )
            y_train, y_test = target.iloc[train_index, :], target.iloc[test_index, :]
            weights_train, weights_test = (
                weights.iloc[train_index],
                weights.iloc[test_index],
            )
            group_train, group_test = (
                groups.iloc[train_index],
                groups.iloc[test_index],
            )

        ### Transform features

        (
            transformer,
            extracted_features_train,
            extracted_features_test,
        ) = benchmark_features_transform(
            X_train,
            y_train,
            X_test,
            group_train,
            group_test,
            feature_eng,
            feature_eng_parameters,
            debug,
        )

        if tabular_model in [
            "lightgbm-gbdt",
            "lightgbm-goss",
            "lightgbm-dart",
            "lightgbm-rf",
            "xgboost-dart",
            "xgboost-gbtree",
            "catboost",
            "lightgbm",
            "xgboost",
        ]:
            ### Train Tabular Models
            reg, pred = benchmark_tree_model(
                extracted_features_train,
                y_train,
                weights_train,
                extracted_features_test,
                y_test,
                weights_test,
                tabular_model,
                tabular_hyper,
                additional_hyper,
                debug,
            )

        if tabular_model in [
            "Numerai-MLP",
            "Numerai-LSTM",
            "tabnet",
        ]:
            ### Train Tabular Models
            reg, pred = benchmark_neural_model(
                extracted_features_train,
                y_train,
                weights_train,
                extracted_features_test,
                y_test,
                weights_test,
                tabular_model,
                tabular_hyper,
                additional_hyper,
                debug,
            )

        ## Convert Prediction output to a dataframe
        pred = pd.DataFrame(pred, index=y_test.index, columns=y_test.columns)

        model_name = "{}_{}_{}".format(tabular_model, feature_eng, model_no)

        parameters[model_name] = {
            "feature_eng": feature_eng_parameters.copy(),
            "tabular": tabular_hyper.copy(),
            "additional": additional_hyper.copy(),
        }

        ### Compute model performance
        model_metrics = dict()
        model_metrics["MSE"] = mean_squared_error(y_test, pred)
        model_performance[model_name] = model_metrics.copy()

        #### Training Parameters
        model_params["feature_columns"] = features.columns
        model_params["target_columns"] = target.columns
        model_params["feature_engineering"] = feature_eng
        model_params["tabular_model"] = tabular_model
        model_params["train_start"] = group_train.iloc[0]
        model_params["train_end"] = group_train.iloc[-1]
        model_params["validation_start"] = group_test.iloc[0]
        model_params["validation_end"] = group_test.iloc[-1]
        model_params["model_name"] = model_name
        parameters[model_name]["model"] = model_params.copy()

        if debug:
            print(parameters[model_name])

        if transformer is not None:
            trained_models[model_name] = {
                "transformer": transformer.data,
                "model": reg,
            }
        else:
            trained_models[model_name] = {
                "transformer": None,
                "model": reg,
            }

        data[model_name] = {
            "prediction": pred,
            "y_test": y_test,
        }

        model_no += 1

    return model_performance, trained_models, data, parameters



import pandas as pd
import numpy as np
import joblib, os, shutil, datetime
import torch

import logging, gc

logger = logging.getLogger("Numerai")


from torch import nn
import torch.nn.functional as F
from torch.utils.data import Dataset, DataLoader, random_split, TensorDataset
from pytorch_lightning import Trainer, LightningModule, seed_everything
from pytorch_lightning.callbacks.early_stopping import EarlyStopping
import cupy as cp


# +
## Tabular Models
# -


class TabularModel:
    def __init__(self, nn_model, config):
        """
        Args:
            nn_model (LightningModule): Neural Networks implmented as a LightningModule
            config (dict): A dictionary which contains the parameters for training NN
        """

        self.nn_model = nn_model
        self.config = config
        seed_everything(config.get("seed", 0), workers=True)

    def train(self, X_train, y_train, X_validate, y_validate):
        self.config["input_shape"] = X_train.shape[1]
        self.config["output_shape"] = y_train.shape[1]

        self.network = self.nn_model(self.config)

        early_stop_callback = EarlyStopping(
            monitor="val_loss",
            min_delta=0.00,
            patience=self.config.get("patience", 5),
            verbose=False,
            mode="min",
        )

        ## Assume X is a DataFrame, assume y is a DataFrame or pd Series
        dataset_train = TensorDataset(
            torch.from_numpy(X_train.values), torch.from_numpy(y_train.values)
        )
        dataloader_train = DataLoader(
            dataset_train,
            batch_size=self.config.get("batch_size", 4096),
            num_workers=0,
        )
        dataset_validate = TensorDataset(
            torch.from_numpy(X_validate.values), torch.from_numpy(y_validate.values)
        )
        dataloader_validate = DataLoader(
            dataset_validate,
            batch_size=self.config.get("batch_size", 4096),
            num_workers=0,
        )

        ## Use GPU if possible
        self.trainer = Trainer(
            accelerator="cuda",
            deterministic=True,
            auto_lr_find=True,
            max_epochs=self.config.get("max_epochs", 3),
            callbacks=[early_stop_callback],
        )

        self.trainer.fit(self.network, dataloader_train, dataloader_validate)

    def predict(self, X):
        self.network.eval()
        with torch.no_grad():
            predictions = self.network(torch.from_numpy(X))
        return predictions.numpy()

    def load_model(self, checkpoint):
        self.network = self.nn_model.load_from_checkpoint(checkpoint)

    def save_model(self, checkpoint):
        self.trainer.save_checkpoint(checkpoint)


# +
## Tabular Modules
# -


class MLP(LightningModule):
    def __init__(self, config):
        super().__init__()
        self.config = config

        neuron_sizes = config.get("neurons", 256)
        num_layers = config.get("num_layers", 2)

        self.layers = nn.Sequential(
            nn.Linear(config["input_shape"], neuron_sizes),
            nn.ReLU(),
            nn.Dropout(config.get("dropout", 0.5)),
        )

        for i in range(
            1,
            num_layers,
        ):
            new_neuron_sizes = int(neuron_sizes * config.get("neuron_scale", 0.5)) + 1
            self.layers.append(
                nn.Linear(neuron_sizes, new_neuron_sizes),
            )
            self.layers.append(nn.ReLU())
            self.layers.append(nn.Dropout(config.get("dropout", 0.5)))
            neuron_sizes = new_neuron_sizes

        self.layers.append(nn.Linear(neuron_sizes, config["output_shape"]))

        ## Need to have this to ensure correct hyper-parameters are loaded
        ## https://github.com/Lightning-AI/lightning/issues/3981
        self.save_hyperparameters()

    def forward(self, x):
        return self.layers(x.float())

    def training_step(self, batch, batch_idx):
        x, y = batch
        y_hat = self.layers(x.float())
        loss = F.mse_loss(y_hat, y.float())
        self.log("train_loss", loss)
        return loss

    def validation_step(self, batch, batch_idx):
        x, y = batch
        y_hat = self.layers(x.float())
        loss = F.mse_loss(y_hat, y.float())
        self.log("val_loss", loss)

    def predict_step(self, batch, batch_idx):
        return self(batch)

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(
            self.parameters(), lr=self.config.get("learning_rate", 1e-4)
        )
        return optimizer


class LSTM_Tabular(LightningModule):
    def __init__(self, config):
        super().__init__()
        self.config = config
        self.lstm = nn.LSTM(
            input_size=self.config.get("no_channels", 4),
            hidden_size=self.config.get("hidden_size", 4),
            num_layers=self.config.get("num_layers", 2),
            dropout=self.config.get("dropout", 0.1),
            batch_first=True,
        )
        self.fc = nn.Linear(
            self.config.get("hidden_size", 4), self.config.get("output_shape", 1)
        )
        self.save_hyperparameters()

    def forward(self, x):
        batch_size, flattened = x.shape
        x = torch.reshape(
            x,
            (
                batch_size,
                -1,
                self.config.get("no_channels", 4),
            ),
        )
        flip_columns_order = self.config.get("flip_column_order", True)
        if flip_columns_order:
            x = torch.flip(x, [1])
        ## LSTM Layers
        lstm_out, _ = self.lstm(
            x.float()
        )  # lstm_out = (batch_size, seq_len, hidden_size)
        x = self.fc(lstm_out[:, -1])
        return x

    def training_step(self, batch, batch_idx):
        x, y = batch
        y_hat = self.forward(x)
        loss = F.mse_loss(y_hat.float(), y.float())
        self.log("train_loss", loss)
        return loss

    def validation_step(self, batch, batch_idx):
        x, y = batch
        y_hat = self.forward(x)
        loss = F.mse_loss(y_hat.float(), y.float())
        self.log("val_loss", loss)

    def predict_step(self, batch, batch_idx):
        return self(batch)

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=1e-4)
        return optimizer


# +
## Time Series Models
## Needs Further Development
# -


class TimeSeriesDataset(Dataset):
    """Face Landmarks dataset."""

    def __init__(self, timeseries, targets=None, lookback=200):
        """
        Args:
            timeseries (pd.DataFrame): A DataFrame of a multivaraite time-series
            targets (pd.DataFrame): A DataFrame of targets for the time-series
            lookback (int): Number of data records to include in lookback
        """
        self.X = timeseries
        self.y = targets
        self.lookback = lookback

    def __len__(self):
        return self.X.shape[0] - (self.lookback - 1)

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()
        if self.y is not None:
            return torch.tensor(
                self.X.values[idx : idx + self.lookback, :]
            ), torch.tensor(self.y.values[idx + self.lookback - 1, :])
        else:
            return torch.tensor(self.X.values[idx : idx + self.lookback, :])


class TimeSeriesModel:
    def __init__(self, nn_model, config):
        """
        Args:
            nn_model (LightningModule): Neural Networks implmented as a LightningModule
            config (dict): A dictionary which contains the parameters for training NN
        """

        self.nn_model = nn_model
        self.config = config
        seed_everything(config.get("seed", 0), workers=True)

    def train(self, X_train, y_train, X_validate, y_validate):
        self.config["input_shape"] = X_train.shape[1]
        self.config["output_shape"] = y_train.shape[1]

        self.network = self.nn_model(self.config)

        early_stop_callback = EarlyStopping(
            monitor="val_loss",
            min_delta=0.00,
            patience=self.config.get("patience", 10),
            verbose=False,
            mode="min",
        )

        ## Assume X is a DataFrame, assume y is a DataFrame or pd Series

        train_dataset = TimeSeriesDataset(
            X_train, y_train, lookback=self.config.get("lookback", 200)
        )
        dataloader_train = torch.utils.data.DataLoader(
            train_dataset, batch_size=self.config.get("batch_size", 1000), shuffle=False
        )
        validate_dataset = TimeSeriesDataset(
            X_validate, y_validate, lookback=self.config.get("lookback", 200)
        )
        dataloader_validate = torch.utils.data.DataLoader(
            validate_dataset,
            batch_size=self.config.get("batch_size", 1000),
            shuffle=False,
        )

        ## Use GPU if possible
        self.trainer = Trainer(
            accelerator="cuda",
            deterministic=True,
            auto_lr_find=True,
            max_epochs=self.config.get("max_epochs", 3),
            callbacks=[early_stop_callback],
        )

        self.trainer.fit(self.network, dataloader_train, dataloader_validate)

    def predict(self, X):
        self.network.eval()
        with torch.no_grad():
            predictions = self.network(X)
        return predictions.numpy()

    def load_model(self, checkpoint):
        self.network = self.nn_model.load_from_checkpoint(checkpoint)

    def save_model(self, checkpoint):
        self.trainer.save_checkpoint(checkpoint)


# +
### TimeSeires Modules
# -


class LSTM(LightningModule):
    def __init__(self, config):
        super().__init__()
        self.config = config
        self.lstm = nn.LSTM(
            input_size=self.config.get("input_size", 11),
            hidden_size=self.config.get("hidden_size", 4),
            num_layers=self.config.get("num_layers", 2),
            dropout=self.config.get("dropout", 0.1),
            batch_first=True,
        )
        self.fc = nn.Linear(self.config.get("hidden_size", 4), 11)

    def forward(self, x):
        lstm_out, _ = self.lstm(x)  # lstm_out = (batch_size, seq_len, hidden_size)
        x = self.fc(lstm_out[:, -1])
        return x

    def training_step(self, batch, batch_idx):
        x, y = batch
        y_hat = self.forward(x.float())
        loss = F.mse_loss(y_hat, y.float())
        self.log("train_loss", loss)
        return loss

    def validation_step(self, batch, batch_idx):
        x, y = batch
        y_hat = self.forward(x.float())
        loss = F.mse_loss(y_hat, y.float())
        self.log("val_loss", loss)

    def predict_step(self, batch, batch_idx):
        return self(batch)

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=1e-4)
        return optimizer



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# A collection of feature enginnering methods for time-series data
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from joblib import Parallel, delayed
import pandas as pd
import numpy as np
from sklearn.base import TransformerMixin, BaseEstimator
import cupy as cp
import cuml
import torch, signatory


import logging

logger = logging.getLogger("Numerai")


"""
Feature Engineering used in Numerai Thesis 
"""


class NumeraiTransformer(TransformerMixin, BaseEstimator):
    def __init__(
        self,
        seed=0,
        usesquare=False,
        dropout_pct=0.05,
        no_product_features=10,
        no_pca_features=0,
    ):
        self.seed = seed
        self.usesquare = usesquare
        self.dropout_pct = dropout_pct
        self.no_product_features = no_product_features
        self.no_pca_features = no_pca_features
        ## Data Dictionary to reconsturct transformer during inference
        self.data = dict()

    ## Transform Numerai Features with mean zero (-2,-1,0,1,2)
    def transform(self, X, is_train=True):
        ## Numpy Random Number Generator
        rng = np.random.default_rng(self.seed)

        ## Drop Out Matrix
        if self.dropout_pct > 0 and is_train:
            dropout_matrix = 1 - np.random.binomial(1, self.dropout_pct, X.shape)
            X_val = X.values * dropout_matrix

        if self.usesquare:
            squareX = pd.DataFrame(np.square(X_val), index=X.index)
            squareX.columns = ["{}_square".format(x) for x in X.columns]
        else:
            squareX = pd.DataFrame()

        ## Pair Transforms
        if self.no_product_features > 0:
            if is_train:
                col1 = np.random.choice(X.columns, self.no_product_features)
                col2 = np.random.choice(
                    X.columns,
                    self.no_product_features,
                )
                self.product_features = pd.DataFrame(
                    {
                        "col1": col1,
                        "col2": col2,
                    }
                ).drop_duplicates()
                self.data["product_features"] = self.product_features
            else:
                self.product_features = self.data["product_features"]

            productX = pd.DataFrame(
                np.array(X[self.product_features["col1"]])
                * np.array(X[self.product_features["col2"]]),
                index=X.index,
            )
            productX.columns = [
                f"feature_product_{i}" for i in range(self.product_features.shape[0])
            ]
        else:
            productX = pd.DataFrame()

        ## Concat All Features to output
        transformed_features = pd.concat(
            [
                X.astype(np.int8),
                squareX.astype(np.int8),
                productX.astype(np.int8),
            ],
            axis=1,
        )

        return transformed_features


class SignatureTransformer(TransformerMixin, BaseEstimator):
    def __init__(
        self,
        lookback,
        signature_level,
    ):
        self.lookback = lookback
        self.signature_level = signature_level

    ## Can also be used to transform data in an online fashion by transform
    def transform(self, X):
        history_length = X.shape[0]
        path_class = signatory.Path(
            torch.Tensor(cp.asarray([X.values])), self.signature_level
        )
        sigs = list()
        for i in range(self.lookback, history_length):
            sigs.append(path_class.logsignature(i - self.lookback, i))
        all_sig = torch.concat(sigs)
        transformed_signature = pd.DataFrame(
            all_sig.numpy(), index=X.index[self.lookback :]
        )
        transformed_signature.columns = [
            "lookback_{}_signature_{}".format(self.lookback, i)
            for i in range(transformed_signature.shape[1])
        ]
        return transformed_signature


def features_transform_batch(transformer, data, is_train=True):
    BATCH_SIZE = 10000000000
    start_index = 0
    transformed_features_batches = list()

    while start_index < data.shape[0]:
        data_batch = data.iloc[start_index : start_index + BATCH_SIZE]
        transformed_featrues_batch = pd.DataFrame(
            transformer.transform(data_batch, is_train=is_train), index=data_batch.index
        )
        transformed_features_batches.append(transformed_featrues_batch)
        start_index = start_index + BATCH_SIZE

    transformed_features = pd.concat(transformed_features_batches, axis=0)
    return transformer, transformed_features


def benchmark_features_transform(
    X_train,
    y_train,
    X_test=None,
    group_train=None,
    group_test=None,
    feature_eng=None,
    feature_eng_parameters=None,
    debug=False,
):
    ### Numerai
    if feature_eng in [
        "numerai",
    ]:
        if feature_eng_parameters is None:
            feature_eng_parameters = {
                "usesquare": False,
                "no_product_features": 0,
                "seed": 10,
            }
        transformer = NumeraiTransformer(**feature_eng_parameters)

    if feature_eng is not None:
        extracted_features_train = transformer.transform(X_train, is_train=True)
        if X_test is not None:
            extracted_features_test = transformer.transform(X_test, is_train=False)
        else:
            extracted_features_test = None

        return transformer, extracted_features_train, extracted_features_test
    else:
        if X_test is not None:
            return None, X_train, X_test
        else:
            return None, X_train, None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# A collection of tools for data pre-processing for non-stationary time-series and tabular data
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import pandas as pd
import numpy as np
import joblib, os, glob

from sklearn.model_selection import TimeSeriesSplit
from sklearn.model_selection._split import _BaseKFold, indexable, _num_samples

from sklearn.inspection import permutation_importance
from sklearn.preprocessing import StandardScaler
from sklearn.base import TransformerMixin, BaseEstimator


"""
Strategy Metrics , Regime Analysis 
"""


def strategy_metrics(strategy, interval=1, numerai=True, accuracy=4):
    results = dict()
    results["mean"] = np.around(strategy.mean(), accuracy)
    results["volatility"] = np.around(strategy.std(), accuracy)
    results["skew"] = np.around(strategy.skew(), accuracy)
    results["kurtosis"] = np.around(strategy.kurtosis(), accuracy)
    if numerai:
        portfolio = strategy.cumsum()
    else:
        portfolio = (1 + strategy).cumprod()
    if numerai:
        dd = portfolio - portfolio.cummax()
    else:
        dd = (portfolio - portfolio.cummax()) / portfolio.cummax()
    results["max_drawdown"] = np.around(-1 * dd.cummin().min(), accuracy).item()
    if strategy.std() > 0:
        results["sharpe"] = np.around(strategy.mean() / strategy.std(), accuracy)
    else:
        results["sharpe"] = np.around(results["mean"] / 1e-4, accuracy)
    if results["max_drawdown"] > 0:
        results["calmar"] = np.around(
            results["mean"] / results["max_drawdown"], accuracy
        )
    else:
        results["calmar"] = np.around(results["mean"] / 1e-4, accuracy)
    return results


def regime_analysis(
    df,
    performance_col="correlation",
    regime_columns="regime",
):
    ans = df.groupby(regime_columns).agg({performance_col: strategy_metrics})
    ans_df = pd.DataFrame(ans[performance_col].values.tolist())
    ans_df.index = ans.index
    return ans_df.reset_index()


"""
Dynamic Model Selection 
"""


def dynamic_model_selection_masks(performances, gap=6, lookback=52, top_models=1):
    mean = performances.shift(gap).rolling(lookback).mean()
    volatility = performances.shift(gap).rolling(lookback).std()
    skew = performances.shift(gap).rolling(lookback).skew()
    kurt = performances.shift(gap).rolling(lookback).kurt()
    drawdown = (
        -1
        * (
            performances.shift(gap).cumsum() - performances.shift(gap).cumsum().cummax()
        ).cummin()
    )
    sharpe = mean / volatility
    calmar = mean / drawdown

    metric_masks = dict()
    for metric in [
        "mean",
        "volatility",
        "skew",
        "kurt",
        "drawdown",
        "sharpe",
        "calmar",
    ]:
        metric_masks[f"{metric}_min"] = np.where(
            locals()[metric].rank(
                axis=1,
                ascending=True,
                na_option="bottom",
            )
            <= top_models,
            1 / top_models,
            np.nan,
        )
        metric_masks[f"{metric}_max"] = np.where(
            locals()[metric].rank(
                axis=1,
                ascending=False,
                na_option="bottom",
            )
            <= top_models,
            1 / top_models,
            np.nan,
        )

    masks_dataframes = dict()
    for metric in [
        "mean",
        "volatility",
        "skew",
        "kurt",
        "drawdown",
        "sharpe",
        "calmar",
    ]:
        masks_dataframes[f"{metric}_min"] = pd.DataFrame(
            metric_masks[f"{metric}_min"],
            columns=locals()[metric].columns,
            index=locals()[metric].index,
        )
        masks_dataframes[f"{metric}_max"] = pd.DataFrame(
            metric_masks[f"{metric}_max"],
            columns=locals()[metric].columns,
            index=locals()[metric].index,
        )
    return masks_dataframes


def walk_forward_dynamic_models(df_list):
    Model_Sets = dict()
    Imputed_Models = dict()

    for key in [
        "Ensemble",
        "Baseline",
        "Optimizer",
        "Small",
        "Medium",
        "Standard",
        "Average",
    ]:
        Model_Sets[key] = list()

    for dynamic_models in df_list:
        Model_Sets["Ensemble"].append(
            dynamic_models[
                [
                    x
                    for x in dynamic_models.columns
                    if "baseline" in x
                    or "optimizer" in x
                    or ("standard" in x and not "average" in x and not "random" in x)
                ]
            ]
        )
        Model_Sets["Average"].append(
            dynamic_models[
                [
                    x
                    for x in dynamic_models.columns
                    if "baseline" in x or "optimizer" in x or "average" in x
                ]
            ]
        )
        Model_Sets["Baseline"].append(
            dynamic_models[[x for x in dynamic_models.columns if "baseline" in x]]
        )
        Model_Sets["Optimizer"].append(
            dynamic_models[[x for x in dynamic_models.columns if "optimizer" in x]]
        )
        Model_Sets["Standard"].append(
            dynamic_models[
                [
                    x
                    for x in dynamic_models.columns
                    if "standard" in x and not "average" in x and not "random" in x
                ]
            ]
        )
        Model_Sets["Small"].append(
            dynamic_models[
                [
                    x
                    for x in dynamic_models.columns
                    if "small" in x and not "average" in x and not "random" in x
                ]
            ]
        )

    for key in [
        "Ensemble",
        "Baseline",
        "Optimizer",
        "Small",
        "Standard",
        "Average",
    ]:
        models_over_time = pd.concat(Model_Sets[key], axis=1)
        # models_over_time = (
        # models_over_time.transpose()
        # .fillna(models_over_time.mean(axis=1))
        # .transpose()
        # )
        models_over_time = models_over_time.transpose().fillna(0).transpose()
        Imputed_Models[key] = models_over_time.sort_index()

    return Imputed_Models


### Compare Against All Trained Models


def create_leaderboard(
    performances_folder,
    searchkey="*",
    lookback=52,
    no_tops=1,
    model_no_lower=0,
    model_no_upper=1e8,
):
    ## Load csv files
    performances_files = sorted(glob.glob(f"{performances_folder}/{searchkey}.csv"))
    models_list = list()
    for f in performances_files:
        model_no = int(f.split(".csv")[0].split("_")[-2])
        model_seq = int(f.split(".csv")[0].split("_")[-1])
        model_name = "_".join(f.split(".csv")[0].split("/")[-1].split("_")[:3])
        if (
            os.path.isfile(f)
            and model_no_lower <= model_no
            and model_no <= model_no_upper
        ):
            df = pd.read_csv(f, index_col=0).sort_index()
            df = df[~df.index.duplicated()]
            df.index = pd.to_datetime(df.index)
            models_list.append(df)

    dynamic_models_collection = walk_forward_dynamic_models(models_list)

    ### Compute Performances of Portfolios of dynamically selected models
    recent_results = list()
    dynamic_portfolios = dict()
    gap = 6
    criteria = [
        "mean",
        # "calmar",
        # "sharpe",
    ]

    for Sets in [
        "Baseline",
        "Optimizer",
        "Ensemble",
        "Small",
        "Standard",
    ]:
        df = dynamic_models_collection[Sets].sort_index()
        if df.shape[0] > 0:
            dynamic_masks = dynamic_model_selection_masks(
                df, top_models=no_tops, lookback=lookback, gap=gap
            )
            for base_method in criteria:
                for method in [
                    f"{base_method}_max",
                ]:
                    portfolio = (dynamic_masks[method] * df).sum(axis=1, min_count=1)
                    dynamic_portfolios[
                        f"{Sets}_{method}_{no_tops}_lookback_{lookback}"
                    ] = portfolio.tail(df.shape[0] - lookback - gap)
                    performances = strategy_metrics(
                        portfolio.tail(df.shape[0] - lookback - gap)
                    )
                    performances["method"] = method
                    performances["no_tops"] = no_tops
                    performances["sets"] = Sets
                    performances["lookback"] = lookback
                    recent_results.append(performances)

    dynamic_performances = pd.DataFrame(recent_results).dropna()

    leaderboards = dict()
    ## Recent Leaderboards to be used in Model Submissions
    for model_subset in [
        "Baseline",
        "Ensemble",
        "Optimizer",
        "Small",
        "Standard",
    ]:
        leaderboard = pd.DataFrame(
            dynamic_models_collection[model_subset]
            .sort_index()
            .iloc[-1 * lookback :]
            .apply(strategy_metrics)
            .to_dict()
        ).transpose()

        if len(dynamic_models_collection[model_subset].columns) > 0:
            leaderboard.index = dynamic_models_collection[model_subset].columns
            leaderboard["proportion"] = [
                float(x[-1]) for x in leaderboard.index.str.split("-")
            ]
            leaderboard["flavour"] = [x[-2] for x in leaderboard.index.str.split("-")]
            leaderboard["model_seq"] = [
                int("-".join(x[:-2]).split("_")[-1])
                for x in leaderboard.index.str.split("-")
            ]
            leaderboard["model_seed"] = [
                int("-".join(x[:-2]).split("_")[-2])
                for x in leaderboard.index.str.split("-")
            ]
            leaderboard["model_cv"] = [
                "-".join(x[:-2]).split("_")[-3]
                for x in leaderboard.index.str.split("-")
            ]
            leaderboard["model_feature_engineering"] = [
                "-".join(x[:-2]).split("_")[-4]
                for x in leaderboard.index.str.split("-")
            ]
            leaderboard["model_tabular_method"] = [
                "-".join(x[:-2]).split("_")[-5]
                for x in leaderboard.index.str.split("-")
            ]
            leaderboards[model_subset] = leaderboard

            ## Leaderboard Since beginning of data
            if dynamic_models_collection[model_subset].shape[0] < lookback + gap:
                start_of_data = 0
            else:
                start_of_data = lookback + gap
            leaderboard = pd.DataFrame(
                dynamic_models_collection[model_subset]
                .sort_index()
                .iloc[start_of_data:]
                .apply(strategy_metrics)
                .to_dict()
            ).transpose()
            leaderboard.index = dynamic_models_collection[model_subset].columns
            leaderboards[f"{model_subset}-All"] = leaderboard

    return (
        dynamic_performances,
        dynamic_portfolios,
        dynamic_models_collection,
        leaderboards,
    )


"""
Cross Validation Schemes

TimeSeries Grouped CV 

"""


class GroupedTimeSeriesSplit(TimeSeriesSplit):
    def __init__(
        self,
        n_splits=5,
        valid_splits=1,
        max_train_size=None,
        test_size=52 * 2,
        gap=52,
        debug=False,
    ):
        self.n_splits = n_splits
        self.valid_splits = valid_splits
        self.shuffle = False
        self.random_state = None
        self.max_train_size = max_train_size
        self.test_size = test_size
        self.gap = gap
        self.debug = debug

    def split(self, X, y=None, groups=None):
        """Generate indices to split data into training and test set.
        Parameters
        ----------
        X : pd.DataFrame of shape (n_samples, n_features)
            Training data, where `n_samples` is the number of samples
            and `n_features` is the number of features.
        y : array-like of shape (n_samples,)
            Always ignored, exists for compatibility.
        groups : pd.Series of shape (n_samples,)
            Group Labels of training data
        Yields
        ------
        train : ndarray
            The training set indices for that split.
        test : ndarray
            The testing set indices for that split.
        """
        X, y, groups = indexable(X, y, groups)
        n_samples = _num_samples(X)

        if groups is None:
            # n_samples = X.shape[0]
            n_splits = self.n_splits
            valid_splits = self.valid_splits
            n_folds = n_splits + 1
            gap = self.gap
            test_size = (
                self.test_size if self.test_size is not None else n_samples // n_folds
            )

            # Make sure we have enough samples for the given split parameters
            if n_folds > n_samples:
                raise ValueError(
                    f"Cannot have number of folds={n_folds} greater"
                    f" than the number of samples={n_samples}."
                )
            if n_samples - gap - (test_size * n_splits) <= 0:
                raise ValueError(
                    f"Too many splits={n_splits} for number of samples"
                    f"={n_samples} with test_size={test_size} and gap={gap}."
                )

            indices = X.index
            test_starts = range(
                n_samples - valid_splits * test_size, n_samples, test_size
            )

            for test_start in test_starts:
                train_end = test_start - gap
                if self.max_train_size and self.max_train_size < train_end:
                    yield (
                        indices[max(train_end - self.max_train_size, 0) : train_end],
                        indices[test_start : test_start + test_size],
                    )
                else:
                    yield (
                        indices[:train_end],
                        indices[test_start : test_start + test_size],
                    )
        else:
            ## Get unique groups
            unique_groups = groups.unique()
            gap = self.gap
            ## Calculate test size if not provided
            if self.test_size:
                n_folds = (len(unique_groups) - gap) // self.test_size
            else:
                n_folds = self.n_splits + 1
                self.test_size = len(unique_groups) // n_folds
            test_splits = [
                unique_groups[
                    len(unique_groups)
                    - (i + 1) * self.test_size : len(unique_groups)
                    - i * self.test_size
                ]
                for i in range(n_folds - 1)
            ]
            if self.max_train_size:
                train_splits = [
                    unique_groups[
                        max(
                            len(unique_groups)
                            - (i + 1) * self.test_size
                            - gap
                            - self.max_train_size,
                            0,
                        ) : len(unique_groups)
                        - (i + 1) * self.test_size
                        - gap
                    ]
                    for i in range(n_folds - 1)
                ]
            else:
                train_splits = [
                    unique_groups[: len(unique_groups) - (i + 1) * self.test_size - gap]
                    for i in range(n_folds - 1)
                ]
            for i in range(0, self.valid_splits):
                yield (
                    groups[groups.isin(train_splits[i])].index,
                    groups[groups.isin(test_splits[i])].index,
                )


"""
Data Dimension Transformer
Currently Implemeted: Constant lookback size with zero-padding
Convert from 2D DataFrame, 
given a lookback size into nested DataFrames for sktime transformers
"""


def forward_fill_zero(series, length):
    fill_length = length - series.shape[0]
    fill_series = pd.Series(np.zeros(fill_length))
    return pd.concat([fill_series, series], axis=0).reset_index(drop=True)


### Create rolling windows of nested dataframe for sktime, forward fill zero if there are not enough data at the start
def roll_2D_to_nested(X, lookback=20, normalise=True):
    ## Python index start at zero
    lookback = lookback - 1
    index = X.index
    columns = X.columns
    output = np.empty((len(index), len(columns)), dtype=object)
    for i in range(X.shape[0]):
        for j, c in enumerate(X.columns):
            start_index = max(0, i - lookback)
            recent_rawdata = pd.Series(X.loc[X.index[start_index : i + 1], c])
            if normalise and i >= 1:
                normalised_rawdata = (
                    recent_rawdata - recent_rawdata.mean()
                ) / recent_rawdata.std()
                output[i, j] = forward_fill_zero(normalised_rawdata, lookback + 1)
            else:
                output[i, j] = forward_fill_zero(recent_rawdata, lookback + 1)
    return pd.DataFrame(output, index=index, columns=columns)


class RollingTSTransformer(BaseEstimator, TransformerMixin):
    def __init__(self, lookback=20, normalise=True):
        self.lookback = lookback
        self.normalise = normalise

    def fit(self, X, y):
        return self

    def transform(self, X):
        output = roll_2D_to_nested(X, self.lookback, self.normalise)
        return output


### Data Pre-processing


def align_features_target(features, target, large_value=1e6):
    ## Flatten multi-index column names for tsfresh
    if isinstance(features, pd.DataFrame):
        if features.columns.nlevels > 1:
            features.columns = [
                "_".join(column).rstrip("_")
                for column in features.columns.to_flat_index()
            ]
    ## Remove rows with na and align features and target to same length
    ##features.replace(np.inf, large_value, inplace=True)
    ##features.replace(-np.inf, -1 * large_value, inplace=True)
    ##features = features.dropna()
    ##target = target.dropna()
    valid_index = features.index.intersection(target.index)
    features = features.reindex(valid_index)
    target = target.reindex(valid_index)
    return features, target
